import re
import sys
from unittest.mock import MagicMock, patch

import pytest
import torch
from datasets import Dataset
from pytest import MonkeyPatch
from torch.testing import assert_close
from transformers.trainer_utils import TrainOutput

from fed_rag.data_collators.huggingface import DataCollatorForLSR
from fed_rag.exceptions import (
    InvalidDataCollatorError,
    InvalidLossError,
    MissingExtraError,
    MissingInputTensor,
    TrainerError,
)
from fed_rag.trainers.huggingface.lsr import (
    HuggingFaceTrainerForLSR,
    LSRSentenceTransformerTrainer,
)
from fed_rag.types.rag_system import RAGSystem


def test_init(
    hf_rag_system: RAGSystem, train_dataset: Dataset, monkeypatch: MonkeyPatch
) -> None:
    # skip validation of rag system
    monkeypatch.setenv("FEDRAG_SKIP_VALIDATION", "1")

    trainer = HuggingFaceTrainerForLSR(
        model=hf_rag_system.retriever.encoder,
        rag_system=hf_rag_system,
        train_dataset=train_dataset,
    )

    assert trainer.train_dataset == train_dataset
    assert trainer.model == hf_rag_system.retriever.encoder
    assert trainer.rag_system == hf_rag_system
    assert isinstance(trainer.hf_trainer_obj, LSRSentenceTransformerTrainer)


def test_invalid_retriever_raises_error(
    mock_rag_system: RAGSystem,
    train_dataset: Dataset,
    monkeypatch: MonkeyPatch,
) -> None:
    # skip validation of rag system
    monkeypatch.setenv("FEDRAG_SKIP_VALIDATION", "1")

    with pytest.raises(
        TrainerError,
        match=(
            "For `HuggingFaceTrainerForLSR`, attribute `model` must be of type "
            "`~sentence_transformers.SentenceTransformer`."
        ),
    ):
        HuggingFaceTrainerForLSR(
            rag_system=mock_rag_system,
            train_dataset=train_dataset,
        )


def test_huggingface_extra_missing(
    train_dataset: Dataset, hf_rag_system: RAGSystem, monkeypatch: MonkeyPatch
) -> None:
    modules = {
        "transformers": None,
    }
    modules_to_import = [
        "fed_rag.trainers.huggingface.mixin",
        "fed_rag.trainers.huggingface.lsr",
    ]
    original_modules = [sys.modules.pop(m, None) for m in modules_to_import]

    with patch.dict("sys.modules", modules):
        msg = (
            "`HuggingFaceTrainerForLSR` requires `huggingface` extra to be installed. "
            "To fix please run `pip install fed-rag[huggingface]`."
        )
        with pytest.raises(
            MissingExtraError,
            match=re.escape(msg),
        ):
            from fed_rag.trainers.huggingface.lsr import (
                HuggingFaceTrainerForLSR,
            )

            HuggingFaceTrainerForLSR(
                rag_system=hf_rag_system,
                train_dataset=train_dataset,
            )

    # restore module so to not affect other tests
    for ix, original_module in enumerate(original_modules):
        if original_module:
            sys.modules[modules_to_import[ix]] = original_module


@patch.object(HuggingFaceTrainerForLSR, "hf_trainer_obj")
def test_train(
    mock_hf_trainer: MagicMock,
    hf_rag_system: RAGSystem,
    train_dataset: Dataset,
    monkeypatch: MonkeyPatch,
) -> None:
    # skip validation of rag system
    monkeypatch.setenv("FEDRAG_SKIP_VALIDATION", "1")

    trainer = HuggingFaceTrainerForLSR(
        rag_system=hf_rag_system,
        train_dataset=train_dataset,
    )
    mock_hf_trainer.train.return_value = TrainOutput(
        global_step=42, training_loss=0.42, metrics={}
    )

    out = trainer.train()

    mock_hf_trainer.train.assert_called_once()
    assert out.loss == 0.42


def test_evaluate(
    hf_rag_system: RAGSystem,
    train_dataset: Dataset,
    monkeypatch: MonkeyPatch,
) -> None:
    # skip validation of rag system
    monkeypatch.setenv("FEDRAG_SKIP_VALIDATION", "1")

    trainer = HuggingFaceTrainerForLSR(
        train_dataset=train_dataset,
        rag_system=hf_rag_system,
    )

    with pytest.raises(NotImplementedError):
        trainer.evaluate()


# test LSRSentenceTransformerTrainer
def test_lsr_sentence_transformer_training_missing_extra_error(
    hf_rag_system: RAGSystem,
    monkeypatch: MonkeyPatch,
) -> None:
    # skip validation of rag system
    monkeypatch.setenv("FEDRAG_SKIP_VALIDATION", "1")

    modules = {
        "sentence_transformers": None,
    }
    modules_to_import = [
        "fed_rag.trainers.huggingface.lsr",
    ]
    original_modules = [sys.modules.pop(m, None) for m in modules_to_import]

    with patch.dict("sys.modules", modules):
        msg = (
            "`LSRSentenceTransformerTrainer` requires `huggingface` extra to be installed. "
            "To fix please run `pip install fed-rag[huggingface]`."
        )
        with pytest.raises(
            MissingExtraError,
            match=re.escape(msg),
        ):
            from fed_rag.trainers.huggingface.lsr import (
                LSRSentenceTransformerTrainer,
            )

            LSRSentenceTransformerTrainer(
                model=hf_rag_system.retriever.encoder,
                data_collator=DataCollatorForLSR(rag_system=hf_rag_system),
            )

    # restore module so to not affect other tests
    for ix, original_module in enumerate(original_modules):
        if original_module:
            sys.modules[modules_to_import[ix]] = original_module


def test_lsr_sentence_transformer_raises_invalid_loss_error(
    hf_rag_system: RAGSystem,
    monkeypatch: MonkeyPatch,
) -> None:
    # skip validation of rag system
    monkeypatch.setenv("FEDRAG_SKIP_VALIDATION", "1")

    with pytest.raises(
        InvalidLossError,
        match="`LSRSentenceTransformerTrainer` must use ~fed_rag.loss.LSRLoss`.",
    ):
        LSRSentenceTransformerTrainer(
            model=hf_rag_system.retriever.encoder,
            data_collator=DataCollatorForLSR(rag_system=hf_rag_system),
            loss="loss",
        )


def test_lsr_sentence_transformer_raises_invalid_collator_error(
    hf_rag_system: RAGSystem,
    monkeypatch: MonkeyPatch,
) -> None:
    # skip validation of rag system
    monkeypatch.setenv("FEDRAG_SKIP_VALIDATION", "1")

    with pytest.raises(
        InvalidDataCollatorError,
        match="`LSRSentenceTransformerTrainer` must use ~fed_rag.data_collators.DataCollatorForLSR`.",
    ):
        LSRSentenceTransformerTrainer(
            model=hf_rag_system.retriever.encoder,
            data_collator="DataCollatorForLSR",
        )


@patch.object(LSRSentenceTransformerTrainer, "collect_scores")
def test_lsr_sentence_transformer_compute_loss(
    mock_collect_scores: MagicMock,
    hf_rag_system: RAGSystem,
    monkeypatch: MonkeyPatch,
) -> None:
    # skip validation of rag system
    monkeypatch.setenv("FEDRAG_SKIP_VALIDATION", "1")

    hf_trainer = LSRSentenceTransformerTrainer(
        model=hf_rag_system.retriever.encoder,
        data_collator=DataCollatorForLSR(rag_system=hf_rag_system),
    )
    mock_loss = MagicMock()
    hf_trainer.loss = mock_loss
    mock_collect_scores.return_value = 0, 0

    # act
    hf_trainer.compute_loss(
        model=hf_trainer.model, inputs={}, return_outputs=True
    )

    # assert
    mock_collect_scores.assert_called_once()
    mock_loss.assert_called_once()


def test_lsr_sentence_transformer_collect_scores(
    hf_rag_system: RAGSystem,
    monkeypatch: MonkeyPatch,
) -> None:
    # skip validation of rag system
    monkeypatch.setenv("FEDRAG_SKIP_VALIDATION", "1")

    hf_trainer = LSRSentenceTransformerTrainer(
        model=hf_rag_system.retriever.encoder,
        data_collator=DataCollatorForLSR(rag_system=hf_rag_system),
    )

    ret, lm = hf_trainer.collect_scores(
        inputs={"lm_scores": torch.ones(5), "retrieval_scores": torch.zeros(5)}
    )

    assert_close(lm, torch.ones(5))
    assert_close(ret, torch.zeros(5))


def test_lsr_sentence_transformer_collect_scores_raises_missing_input_tensor_error(
    hf_rag_system: RAGSystem,
    monkeypatch: MonkeyPatch,
) -> None:
    # skip validation of rag system
    monkeypatch.setenv("FEDRAG_SKIP_VALIDATION", "1")

    hf_trainer = LSRSentenceTransformerTrainer(
        model=hf_rag_system.retriever.encoder,
        data_collator=DataCollatorForLSR(rag_system=hf_rag_system),
    )

    with pytest.raises(
        MissingInputTensor,
        match="Collated `inputs` are missing key `retrieval_scores`",
    ):
        hf_trainer.collect_scores(inputs={"lm_scores": torch.ones(5)})

    with pytest.raises(
        MissingInputTensor,
        match="Collated `inputs` are missing key `lm_scores`",
    ):
        hf_trainer.collect_scores(inputs={"retrieval_scores": torch.ones(5)})



"""HuggingFaceFLTask Unit Tests"""

import re
import sys
from typing import Callable, OrderedDict
from unittest.mock import MagicMock, patch

import pytest
import torch
from datasets import Dataset
from flwr.common.parameter import ndarrays_to_parameters
from flwr.server.client_manager import SimpleClientManager
from flwr.server.strategy import FedAvg
from peft import PeftModel
from peft.utils import get_peft_model_state_dict
from torch.nn import Module
from transformers import PreTrainedModel

from fed_rag.exceptions import (
    MissingExtraError,
    MissingRequiredNetParam,
    NetTypeMismatch,
)
from fed_rag.fl_tasks.huggingface import (
    BaseFLTaskBundle,
    HuggingFaceFlowerClient,
    HuggingFaceFLTask,
    MissingTesterSpec,
    MissingTrainerSpec,
    UnequalNetParamWarning,
    _get_weights,
)
from fed_rag.types import TestResult, TrainResult


# TODO: parametrize tests using trainer/tester from pretrained models
def test_init_flower_client(
    train_dataset: Dataset,
    val_dataset: Dataset,
    trainer_pretrained_model: Callable,
    tester_pretrained_model: Callable,
    hf_pretrained_model: PreTrainedModel,
) -> None:
    bundle = BaseFLTaskBundle(
        net=hf_pretrained_model,
        train_dataset=train_dataset,
        val_dataset=val_dataset,
        trainer=trainer_pretrained_model,
        tester=tester_pretrained_model,
        extra_test_kwargs={},
        extra_train_kwargs={},
    )
    client = HuggingFaceFlowerClient(task_bundle=bundle)

    assert client.tester == tester_pretrained_model
    assert client.trainer == trainer_pretrained_model
    assert client.train_dataset == train_dataset
    assert client.val_dataset == val_dataset
    assert client.extra_train_kwargs == {}
    assert client.extra_test_kwargs == {}
    assert client.task_bundle == bundle


def test_flower_client_get_weights(
    train_dataset: Dataset,
    val_dataset: Dataset,
    trainer_pretrained_model: Callable,
    tester_pretrained_model: Callable,
    hf_pretrained_model: PreTrainedModel,
) -> None:
    net = hf_pretrained_model
    bundle = BaseFLTaskBundle(
        net=hf_pretrained_model,
        train_dataset=train_dataset,
        val_dataset=val_dataset,
        trainer=trainer_pretrained_model,
        tester=tester_pretrained_model,
        extra_test_kwargs={},
        extra_train_kwargs={},
    )
    client = HuggingFaceFlowerClient(task_bundle=bundle)
    expected_weights = [
        val.cpu().numpy() for _, val in net.state_dict().items()
    ]

    assert all((client.get_weights()[0] == expected_weights[0]).flatten())
    assert all((client.get_weights()[1] == expected_weights[1]).flatten())


@patch.object(Module, "load_state_dict")
def test_flower_client_set_weights(
    mock_load_state_dict: MagicMock,
    train_dataset: Dataset,
    val_dataset: Dataset,
    trainer_pretrained_model: Callable,
    tester_pretrained_model: Callable,
    hf_pretrained_model: PreTrainedModel,
) -> None:
    net = hf_pretrained_model
    bundle = BaseFLTaskBundle(
        net=net,
        train_dataset=train_dataset,
        val_dataset=val_dataset,
        trainer=trainer_pretrained_model,
        tester=tester_pretrained_model,
        extra_test_kwargs={},
        extra_train_kwargs={},
    )
    client = HuggingFaceFlowerClient(task_bundle=bundle)
    parameters = client.get_weights()

    # act
    client.set_weights(parameters)

    # assert
    mock_load_state_dict.assert_called_once()
    args, kwargs = mock_load_state_dict.call_args
    params_dict = zip(net.state_dict().keys(), parameters)
    state_dict = OrderedDict({k: torch.tensor(v) for k, v in params_dict})
    assert (args[0].get("model.weight") == state_dict["model.weight"]).all()
    assert (args[0].get("model.bias") == state_dict["model.bias"]).all()
    assert kwargs == {"strict": True}
    assert client.task_bundle == bundle


def test_flower_client_get_weights_peft(
    train_dataset: Dataset,
    val_dataset: Dataset,
    trainer_peft_model: Callable,
    tester_peft_model: Callable,
    hf_peft_model: PeftModel,
) -> None:
    net = hf_peft_model
    bundle = BaseFLTaskBundle(
        net=hf_peft_model,
        train_dataset=train_dataset,
        val_dataset=val_dataset,
        trainer=trainer_peft_model,
        tester=tester_peft_model,
        extra_test_kwargs={},
        extra_train_kwargs={},
    )
    client = HuggingFaceFlowerClient(task_bundle=bundle)
    state_dict = get_peft_model_state_dict(net)
    expected_weights = [val.cpu().numpy() for _, val in state_dict.items()]

    assert all((client.get_weights()[0] == expected_weights[0]).flatten())
    assert all((client.get_weights()[1] == expected_weights[1]).flatten())


@patch("fed_rag.fl_tasks.huggingface.set_peft_model_state_dict")
def test_flower_client_set_weights_peft(
    mock_peft_model_state_dict: MagicMock,
    train_dataset: Dataset,
    val_dataset: Dataset,
    trainer_peft_model: Callable,
    tester_peft_model: Callable,
    hf_peft_model: PeftModel,
) -> None:
    net = hf_peft_model
    bundle = BaseFLTaskBundle(
        net=net,
        train_dataset=train_dataset,
        val_dataset=val_dataset,
        trainer=trainer_peft_model,
        tester=tester_peft_model,
        extra_test_kwargs={},
        extra_train_kwargs={},
    )
    client = HuggingFaceFlowerClient(task_bundle=bundle)
    parameters = client.get_weights()

    # act
    client.set_weights(parameters)

    # assert
    params_dict = zip(get_peft_model_state_dict(net).keys(), parameters)
    expected_state_dict = OrderedDict(
        {k: torch.tensor(v) for k, v in params_dict}
    )
    mock_peft_model_state_dict.assert_called_once()
    (peft_arg, state_dict_arg), _ = mock_peft_model_state_dict.call_args
    assert peft_arg == net
    assert all(
        x[1].equal(y[1])
        for x, y in zip(state_dict_arg.items(), expected_state_dict.items())
    )
    assert client.task_bundle == bundle


def test_flower_client_fit(
    train_dataset: Dataset,
    val_dataset: Dataset,
    trainer_pretrained_model: Callable,
    tester_pretrained_model: Callable,
    hf_pretrained_model: PreTrainedModel,
) -> None:
    net = hf_pretrained_model
    bundle = BaseFLTaskBundle(
        net=net,
        train_dataset=train_dataset,
        val_dataset=val_dataset,
        trainer=trainer_pretrained_model,
        tester=tester_pretrained_model,
        extra_test_kwargs={},
        extra_train_kwargs={},
    )
    client = HuggingFaceFlowerClient(task_bundle=bundle)
    parameters = client.get_weights()
    mock_trainer = MagicMock()
    client.trainer = mock_trainer
    mock_trainer.return_value = TrainResult(loss=0.01)

    # act
    result = client.fit(parameters, config={})

    # assert
    mock_trainer.assert_called_once()
    for a, b in zip(result[0], client.get_weights()):
        assert (a == b).all()
    assert result[1] == len(client.train_dataset)
    assert result[2] == {"loss": 0.01}


def test_flower_client_evaluate(
    train_dataset: Dataset,
    val_dataset: Dataset,
    trainer_pretrained_model: Callable,
    tester_pretrained_model: Callable,
    hf_pretrained_model: PreTrainedModel,
) -> None:
    net = hf_pretrained_model
    bundle = BaseFLTaskBundle(
        net=net,
        train_dataset=train_dataset,
        val_dataset=val_dataset,
        trainer=trainer_pretrained_model,
        tester=tester_pretrained_model,
        extra_test_kwargs={},
        extra_train_kwargs={},
    )
    client = HuggingFaceFlowerClient(task_bundle=bundle)
    parameters = client.get_weights()
    mock_tester = MagicMock()
    client.tester = mock_tester
    mock_tester.return_value = TestResult(
        loss=0.01, metrics={"accuracy": 0.88}
    )

    # act
    result = client.evaluate(parameters=parameters, config={})

    # assert
    mock_tester.assert_called_once()
    assert result[0] == 0.01
    assert result[1] == len(client.val_dataset)
    assert result[2] == {"accuracy": 0.88}


def test_init_from_trainer_tester(
    trainer_pretrained_model: Callable,
    tester_pretrained_model: Callable,
) -> None:
    fl_task = HuggingFaceFLTask.from_trainer_and_tester(
        trainer=trainer_pretrained_model, tester=tester_pretrained_model
    )

    assert fl_task._trainer_spec == getattr(
        trainer_pretrained_model, "__fl_task_trainer_config"
    )
    assert fl_task._tester_spec == getattr(
        tester_pretrained_model, "__fl_task_tester_config"
    )
    assert fl_task._tester == tester_pretrained_model
    assert fl_task._trainer == trainer_pretrained_model
    assert fl_task.training_loop == trainer_pretrained_model


def test_invoking_server_without_strategy_and_net_param_raises(
    trainer_pretrained_model: Callable,
    tester_pretrained_model: Callable,
) -> None:
    fl_task = HuggingFaceFLTask.from_trainer_and_tester(
        trainer=trainer_pretrained_model, tester=tester_pretrained_model
    )
    with pytest.raises(
        MissingRequiredNetParam,
        match="Please pass in a model using the model param name net.",
    ):
        client_manager = SimpleClientManager()
        fl_task.server(client_manager=client_manager)


def test_invoking_server(
    trainer_pretrained_model: Callable,
    tester_pretrained_model: Callable,
) -> None:
    fl_task = HuggingFaceFLTask.from_trainer_and_tester(
        trainer=trainer_pretrained_model, tester=tester_pretrained_model
    )
    strategy = FedAvg()
    client_manager = SimpleClientManager()
    server = fl_task.server(strategy=strategy, client_manager=client_manager)

    assert server.client_manager() == client_manager
    assert server.strategy == strategy


def test_huggingface_extra_missing_fltask(
    trainer_pretrained_model: Callable,
    tester_pretrained_model: Callable,
) -> None:
    """Test extra is not installed.

    Combination of:
    - https://docs.python.org/3/library/unittest.mock.html#order-of-precedence-of-side-effect-return-value-and-wraps
    - https://medium.com/python-pandemonium/how-to-test-your-imports-1461c1113be1
    """

    modules = {"peft": None}
    module_to_import = "fed_rag.fl_tasks.huggingface"

    if module_to_import in sys.modules:
        original_module = sys.modules.pop(module_to_import)

    with patch.dict("sys.modules", modules):
        msg = (
            "`HuggingFaceFLTask` requires `huggingface` extra to be installed. "
            "To fix please run `pip install fed-rag[huggingface]`."
        )
        with pytest.raises(
            MissingExtraError,
            match=re.escape(msg),
        ):
            from fed_rag.fl_tasks.huggingface import HuggingFaceFLTask

            HuggingFaceFLTask.from_trainer_and_tester(
                trainer=trainer_pretrained_model,
                tester=tester_pretrained_model,
            )

    # restore module so to not affect other tests
    sys.modules[module_to_import] = original_module


def test_huggingface_extra_missing_fltask_bundle(
    train_dataset: Dataset,
    val_dataset: Dataset,
    trainer_pretrained_model: Callable,
    tester_pretrained_model: Callable,
    hf_pretrained_model: PreTrainedModel,
) -> None:
    """Test extra is not installed."""

    modules = {"peft": None}
    module_to_import = "fed_rag.fl_tasks.huggingface"

    if module_to_import in sys.modules:
        original_module = sys.modules.pop(module_to_import)

    with patch.dict("sys.modules", modules):
        msg = (
            "`BaseFLTaskBundle` requires `huggingface` extra to be installed. "
            "To fix please run `pip install fed-rag[huggingface]`."
        )
        with pytest.raises(
            MissingExtraError,
            match=re.escape(msg),
        ):
            from fed_rag.fl_tasks.huggingface import BaseFLTaskBundle

            net = hf_pretrained_model
            BaseFLTaskBundle(
                net=net,
                train_dataset=train_dataset,
                val_dataset=val_dataset,
                trainer=trainer_pretrained_model,
                tester=tester_pretrained_model,
                extra_test_kwargs={},
                extra_train_kwargs={},
            )

    # restore module so to not affect other tests
    sys.modules[module_to_import] = original_module


def test_invoking_server_using_defaults(
    trainer_pretrained_model: Callable,
    tester_pretrained_model: Callable,
    hf_pretrained_model: PreTrainedModel,
) -> None:
    fl_task = HuggingFaceFLTask.from_trainer_and_tester(
        trainer=trainer_pretrained_model, tester=tester_pretrained_model
    )
    ndarrays = _get_weights(hf_pretrained_model)
    parameters = ndarrays_to_parameters(ndarrays)

    # act
    server = fl_task.server(net=hf_pretrained_model)

    assert type(server.client_manager()) is SimpleClientManager
    assert type(server.strategy) is FedAvg
    assert server.strategy.initial_parameters == parameters


def test_invoking_client_without_net_param_raises(
    trainer_pretrained_model: Callable,
    tester_pretrained_model: Callable,
) -> None:
    fl_task = HuggingFaceFLTask.from_trainer_and_tester(
        trainer=trainer_pretrained_model, tester=tester_pretrained_model
    )
    with pytest.raises(
        MissingRequiredNetParam,
        match="Please pass in a model using the model param name net.",
    ):
        fl_task.client()


def test_creating_fl_task_with_undecorated_trainer_raises_error(
    undecorated_trainer: Callable,
    tester_pretrained_model: Callable,
) -> None:
    with pytest.raises(
        MissingTrainerSpec,
        match="Cannot extract `TrainerSignatureSpec` from supplied `trainer`.",
    ):
        HuggingFaceFLTask.from_trainer_and_tester(
            trainer=undecorated_trainer,
            tester=tester_pretrained_model,
        )


def test_creating_fl_task_with_undecorated_tested_raises_error(
    trainer_pretrained_model: Callable,
    undecorated_tester: Callable,
) -> None:
    with pytest.raises(
        MissingTesterSpec,
        match="Cannot extract `TesterSignatureSpec` from supplied `tester`.",
    ):
        HuggingFaceFLTask.from_trainer_and_tester(
            trainer=trainer_pretrained_model,
            tester=undecorated_tester,
        )


def test_creating_fl_task_with_mismatched_net_params_raises_warning(
    trainer_pretrained_model: Callable,
    mismatch_tester_pretrained_model: Callable,
) -> None:
    msg = (
        "`trainer`'s model parameter name is not the same as that for `tester`. "
        "Will use the name supplied in `trainer`."
    )
    with pytest.warns(UnequalNetParamWarning, match=msg):
        HuggingFaceFLTask.from_trainer_and_tester(
            trainer=trainer_pretrained_model,
            tester=mismatch_tester_pretrained_model,
        )


def test_init_fl_task_with_mismatched_net_params_raises_warning(
    trainer_pretrained_model: Callable,
    mismatch_tester_pretrained_model: Callable,
) -> None:
    msg = (
        "`trainer`'s model parameter name is not the same as that for `tester`. "
        "Will use the name supplied in `trainer`."
    )
    with pytest.warns(UnequalNetParamWarning, match=msg):
        HuggingFaceFLTask(
            trainer=trainer_pretrained_model,
            trainer_spec=trainer_pretrained_model.__fl_task_trainer_config,  # type: ignore[attr-defined]
            tester=mismatch_tester_pretrained_model,
            tester_spec=mismatch_tester_pretrained_model.__fl_task_tester_config,  # type: ignore[attr-defined]
        )


def test_init_fl_task_with_mismatched_net_param_type_raises_error_i(
    trainer_pretrained_model: Callable,
    tester_sentence_transformer: Callable,
) -> None:
    msg = "`trainer`'s model class is not the same as that for `tester`."
    with pytest.raises(NetTypeMismatch, match=msg):
        HuggingFaceFLTask(
            trainer=trainer_pretrained_model,
            trainer_spec=trainer_pretrained_model.__fl_task_trainer_config,  # type: ignore[attr-defined]
            tester=tester_sentence_transformer,
            tester_spec=tester_sentence_transformer.__fl_task_tester_config,  # type: ignore[attr-defined]
        )


def test_init_fl_task_with_mismatched_net_param_type_raises_error_ii(
    trainer_pretrained_model: Callable,
    tester_peft_model: Callable,
) -> None:
    msg = "`trainer`'s model class is not the same as that for `tester`."
    with pytest.raises(NetTypeMismatch, match=msg):
        HuggingFaceFLTask(
            trainer=trainer_pretrained_model,
            trainer_spec=trainer_pretrained_model.__fl_task_trainer_config,  # type: ignore[attr-defined]
            tester=tester_peft_model,
            tester_spec=tester_peft_model.__fl_task_tester_config,  # type: ignore[attr-defined]
        )


def test_fl_task_methods_not_implemented(
    trainer_pretrained_model: Callable,
    tester_pretrained_model: Callable,
) -> None:
    # from_configs
    with pytest.raises(NotImplementedError):
        HuggingFaceFLTask.from_configs(None, None)

    # simulate
    fl_task = HuggingFaceFLTask.from_trainer_and_tester(
        trainer=trainer_pretrained_model, tester=tester_pretrained_model
    )
    with pytest.raises(NotImplementedError):
        fl_task.simulate(42)


def test_private_get_weights_raises_error_with_invalid_net_type() -> None:
    # act
    with pytest.raises(ValueError):
        _get_weights(3)



"""Qdrant Knowledge Store"""

import warnings
from contextlib import contextmanager
from typing import TYPE_CHECKING, Any, Generator, Literal

from pydantic import Field, SecretStr, model_validator

from fed_rag.base.knowledge_store import BaseKnowledgeStore
from fed_rag.exceptions import (
    InvalidDistanceError,
    KnowledgeStoreError,
    KnowledgeStoreNotFoundError,
    KnowledgeStoreWarning,
    LoadNodeError,
)
from fed_rag.knowledge_stores.qdrant.utils import check_qdrant_installed
from fed_rag.types.knowledge_node import KnowledgeNode

if TYPE_CHECKING:  # pragma: no cover
    from qdrant_client import QdrantClient
    from qdrant_client.http.models import ScoredPoint
    from qdrant_client.models import PointStruct


def _get_qdrant_client(
    host: str,
    port: int,
    grpc_port: int,
    https: bool = False,
    timeout: int | None = None,
    api_key: str | None = None,
    **kwargs: Any,
) -> "QdrantClient":
    """Get a QdrantClient

    NOTE: should be used within `QdrantKnowledgeStore` post validation that the
    qdrant extra has been installed.
    """
    from qdrant_client import QdrantClient

    return QdrantClient(
        host=host,
        port=port,
        grpc_port=grpc_port,
        api_key=api_key,
        timeout=timeout,
        https=https,
        **kwargs,
    )


def _convert_knowledge_node_to_qdrant_point(
    node: KnowledgeNode,
) -> "PointStruct":
    from qdrant_client.models import PointStruct

    return PointStruct(
        id=node.node_id, vector=node.embedding, payload=node.model_dump()
    )


def _convert_scored_point_to_knowledge_node_and_score_tuple(
    scored_point: "ScoredPoint",
) -> tuple[float, KnowledgeNode]:
    return (
        scored_point.score,
        KnowledgeNode.model_validate(scored_point.payload),
    )


class QdrantKnowledgeStore(BaseKnowledgeStore):
    """Qdrant Knowledge Store Class

    NOTE: This is a minimal implementation in order to just get started using Qdrant.
    """

    host: str = Field(default="localhost")
    port: int = Field(default=6333)
    grpc_port: int = Field(default=6334)
    https: bool = Field(default=False)
    api_key: SecretStr | None = Field(default=None)
    collection_name: str = Field(description="Name of Qdrant collection")
    collection_distance: Literal[
        "Cosine", "Euclid", "Dot", "Manhattan"
    ] = Field(
        description="Distance definition for collection", default="Cosine"
    )
    client_kwargs: dict[str, Any] = Field(default_factory=dict)
    timeout: int | None = Field(default=None)
    load_nodes_kwargs: dict[str, Any] = Field(default_factory=dict)

    @contextmanager
    def get_client(
        self,
    ) -> Generator["QdrantClient", None, None]:
        client = _get_qdrant_client(
            host=self.host,
            port=self.port,
            grpc_port=self.grpc_port,
            https=self.https,
            timeout=self.timeout,
            api_key=self.api_key.get_secret_value() if self.api_key else None,
            **self.client_kwargs,
        )

        try:
            yield client
        finally:
            try:
                client.close()
            except Exception as e:
                warnings.warn(
                    f"Unable to close client: {str(e)}", KnowledgeStoreWarning
                )

    def _collection_exists(self) -> bool:
        """Check if a collection exists."""
        with self.get_client() as client:
            return client.collection_exists(self.collection_name)  # type: ignore[no-any-return]

    def _create_collection(
        self, collection_name: str, vector_size: int, distance: str
    ) -> None:
        from qdrant_client.models import Distance, VectorParams

        try:
            # Try to convert to enum
            distance = Distance(distance)
        except ValueError:
            # Catch the ValueError from enum conversion and raise your custom error
            raise InvalidDistanceError(
                f"Unsupported distance: {distance}. "
                f"Mode must be one of: {', '.join([m.value for m in Distance])}"
            )

        with self.get_client() as client:
            try:
                client.create_collection(
                    collection_name=collection_name,
                    vectors_config=VectorParams(
                        size=vector_size, distance=distance
                    ),
                )
            except Exception as e:
                raise KnowledgeStoreError(
                    f"Failed to create collection: {str(e)}"
                ) from e

    def _ensure_collection_exists(self) -> None:
        if not self._collection_exists():
            raise KnowledgeStoreNotFoundError(
                f"Collection '{self.collection_name}' does not exist."
            )

    def _check_if_collection_exists_otherwise_create_one(
        self, vector_size: int
    ) -> None:
        if not self._collection_exists():
            try:
                self._create_collection(
                    collection_name=self.collection_name,
                    vector_size=vector_size,
                    distance=self.collection_distance,
                )
            except Exception as e:
                raise KnowledgeStoreError(
                    f"Failed to create new collection: '{self.collection_name}'"
                ) from e

    @model_validator(mode="before")
    @classmethod
    def check_dependencies(cls, data: Any) -> Any:
        """Validate that qdrant dependencies are installed."""
        check_qdrant_installed()
        return data

    def load_node(self, node: KnowledgeNode) -> None:
        self._check_if_collection_exists_otherwise_create_one(
            vector_size=len(node.embedding)
        )

        point = _convert_knowledge_node_to_qdrant_point(node)
        with self.get_client() as client:
            try:
                client.upsert(
                    collection_name=self.collection_name, points=[point]
                )
            except Exception as e:
                raise LoadNodeError(
                    f"Failed to load node {node.node_id} into collection '{self.collection_name}': {str(e)}"
                ) from e

    def load_nodes(self, nodes: list[KnowledgeNode]) -> None:
        if not nodes:
            return

        self._check_if_collection_exists_otherwise_create_one(
            vector_size=len(nodes[0].embedding)
        )

        points = [_convert_knowledge_node_to_qdrant_point(n) for n in nodes]
        with self.get_client() as client:
            try:
                client.upload_points(
                    collection_name=self.collection_name,
                    points=points,
                    **self.load_nodes_kwargs,
                )
            except Exception as e:
                raise LoadNodeError(
                    f"Loading nodes into collection '{self.collection_name}' failed: {str(e)}"
                ) from e

    def retrieve(
        self, query_emb: list[float], top_k: int
    ) -> list[tuple[float, KnowledgeNode]]:
        """Retrieve top-k nodes from the vector store."""
        from qdrant_client.conversions.common_types import QueryResponse

        self._ensure_collection_exists()

        with self.get_client() as client:
            try:
                hits: QueryResponse = client.query_points(
                    collection_name=self.collection_name,
                    query=query_emb,
                    limit=top_k,
                )
            except Exception as e:
                raise KnowledgeStoreError(
                    f"Failed to retrieve from collection '{self.collection_name}': {str(e)}"
                ) from e

        return [
            _convert_scored_point_to_knowledge_node_and_score_tuple(pt)
            for pt in hits.points
        ]

    def delete_node(self, node_id: str) -> bool:
        """Delete a node based on its node_id."""
        from qdrant_client.http.models import (
            FieldCondition,
            Filter,
            MatchValue,
            UpdateResult,
            UpdateStatus,
        )

        self._ensure_collection_exists()

        with self.get_client() as client:
            try:
                res: UpdateResult = client.delete(
                    collection_name=self.collection_name,
                    points_selector=Filter(
                        must=[
                            FieldCondition(
                                key="node_id", match=MatchValue(value=node_id)
                            )
                        ]
                    ),
                )
            except Exception:
                raise KnowledgeStoreError(
                    f"Failed to delete node: '{node_id}' from collection '{self.collection_name}'"
                )

        return bool(res.status == UpdateStatus.COMPLETED)

    def clear(self) -> None:
        self._ensure_collection_exists()

        # delete the collection
        with self.get_client() as client:
            try:
                client.delete_collection(collection_name=self.collection_name)
            except Exception as e:
                raise KnowledgeStoreError(
                    f"Failed to delete collection '{self.collection_name}': {str(e)}"
                ) from e

    @property
    def count(self) -> int:
        from qdrant_client.http.models import CountResult

        self._ensure_collection_exists()

        with self.get_client() as client:
            try:
                res: CountResult = client.count(
                    collection_name=self.collection_name
                )
                return int(res.count)
            except Exception as e:
                raise KnowledgeStoreError(
                    f"Failed to get vector count for collection '{self.collection_name}': {str(e)}"
                ) from e

    def persist(self) -> None:
        """Persist a knowledge store to disk."""
        raise NotImplementedError(
            "`persist()` is not available in QdrantKnowledgeStore."
        )

    def load(self) -> None:
        """Load a previously persisted knowledge store."""
        raise NotImplementedError(
            "`load()` is not available in QdrantKnowledgeStore. "
            "Data is automatically persisted and loaded from the Qdrant server."
        )



from typing import Any

import pytest
import torch
from pydantic import PrivateAttr
from sentence_transformers import SentenceTransformer

from fed_rag.base.retriever import BaseRetriever


class MockRetriever(BaseRetriever):
    _encoder: torch.nn.Module = PrivateAttr(default=torch.nn.Linear(2, 1))

    def encode_context(self, context: str, **kwargs: Any) -> torch.Tensor:
        return self._encoder.forward(torch.ones(2))

    def encode_query(self, query: str, **kwargs: Any) -> torch.Tensor:
        return self._encoder.forward(torch.zeros(2))

    @property
    def encoder(self) -> torch.nn.Module:
        return self._encoder

    @property
    def query_encoder(self) -> torch.nn.Module | None:
        return None

    @property
    def context_encoder(self) -> torch.nn.Module | None:
        return None


@pytest.fixture
def mock_retriever() -> MockRetriever:
    return MockRetriever()


@pytest.fixture
def dummy_sentence_transformer() -> SentenceTransformer:
    return SentenceTransformer(modules=[torch.nn.Linear(5, 5)])



import re
import sys
from typing import Sequence
from unittest.mock import MagicMock, patch

import pytest
import torch
from pytest import MonkeyPatch
from torch.testing import assert_close

from fed_rag.base.tokenizer import EncodeResult
from fed_rag.data_collators.huggingface.ralt import (
    DEFAULT_EXAMPLE_TEMPLATE,
    DataCollatorForRALT,
)
from fed_rag.exceptions import (
    DataCollatorError,
    FedRAGError,
    MissingExtraError,
)
from fed_rag.generators.huggingface import HFPeftModelGenerator
from fed_rag.retrievers.huggingface.hf_sentence_transformer import (
    HFSentenceTransformerRetriever,
)
from fed_rag.types.knowledge_node import KnowledgeNode
from fed_rag.types.rag_system import RAGSystem, SourceNode


def test_huggingface_extra_missing(mock_rag_system: RAGSystem) -> None:
    """Test extra is not installed."""

    modules = {
        "transformers": None,
        "transformers.data": None,
        "transformers.data.data_collator": None,
    }
    module_to_import = "fed_rag.data_collators.huggingface.ralt"
    original_module = sys.modules.pop(module_to_import, None)

    with patch.dict("sys.modules", modules):
        msg = (
            "`DataCollatorForRALT` requires `huggingface` extra to be installed. "
            "To fix please run `pip install fed-rag[huggingface]`."
        )
        with pytest.raises(
            MissingExtraError,
            match=re.escape(msg),
        ):
            from fed_rag.data_collators.huggingface.ralt import (
                DataCollatorForRALT,
            )

            DataCollatorForRALT(
                rag_system=mock_rag_system,
            )

    # restore module so to not affect other tests
    if original_module:
        sys.modules[module_to_import] = original_module


def test_huggingface_extra_missing_from_parent(
    mock_rag_system: RAGSystem,
) -> None:
    """Test extra is not installed."""

    modules = {
        "transformers": None,
        "transformers.data": None,
        "transformers.data.data_collator": None,
    }
    module_to_import = "fed_rag.data_collators.huggingface"
    original_module = sys.modules.pop(module_to_import, None)

    with patch.dict("sys.modules", modules):
        msg = (
            "`fed_rag.data_collators.huggingface` requires `huggingface` extra to be installed. "
            "To fix please run `pip install fed-rag[huggingface]`."
        )
        with pytest.raises(
            MissingExtraError,
            match=re.escape(msg),
        ):
            from fed_rag.data_collators.huggingface import DataCollatorForRALT

            DataCollatorForRALT(
                rag_system=mock_rag_system,
                prompt_template="{query} and {context}",
            )

    # restore module so to not affect other tests
    if original_module:
        sys.modules[module_to_import] = original_module


def test_invalid_rag_system_due_to_generators(
    mock_rag_system: RAGSystem,
) -> None:
    with pytest.raises(
        FedRAGError,
        match="Generator must be HFPretrainedModelGenerator or HFPeftModelGenerator",
    ):
        from fed_rag.data_collators.huggingface import DataCollatorForRALT

        DataCollatorForRALT(rag_system=mock_rag_system)


def test_invalid_rag_system_due_to_retriever(
    mock_rag_system: RAGSystem,
) -> None:
    generator = HFPeftModelGenerator(
        model_name="fake_name",
        base_model_name="fake_base_name",
        load_model_at_init=False,
    )
    mock_rag_system.generator = generator

    with pytest.raises(
        FedRAGError,
        match="Retriever must be a HFSentenceTransformerRetriever",
    ):
        from fed_rag.data_collators.huggingface import DataCollatorForRALT

        DataCollatorForRALT(rag_system=mock_rag_system, prompt_template="")


def test_invalid_return_tensors(
    mock_rag_system: RAGSystem,
    monkeypatch: MonkeyPatch,
) -> None:
    monkeypatch.setenv("FEDRAG_SKIP_VALIDATION", "1")
    return_tensors = "af"
    with pytest.raises(
        FedRAGError,
        match=f"Framework '{return_tensors}' not recognized!",
    ):
        collator = DataCollatorForRALT(
            rag_system=mock_rag_system, prompt_template=""
        )
        collator([], return_tensors)


def test_init(
    mock_rag_system: RAGSystem,
) -> None:
    generator = HFPeftModelGenerator(
        model_name="fake_name",
        base_model_name="fake_base_name",
        load_model_at_init=False,
    )
    retriever = HFSentenceTransformerRetriever(
        model_name="fake_name", load_model_at_init=False
    )
    mock_rag_system.generator = generator
    mock_rag_system.retriever = retriever

    collator = DataCollatorForRALT(
        rag_system=mock_rag_system,
    )

    assert collator.rag_system == mock_rag_system
    assert collator.example_template == DEFAULT_EXAMPLE_TEMPLATE
    assert collator.default_return_tensors == "pt"


@pytest.fixture()
def mock_examples() -> Sequence[dict]:
    return [
        {
            "query": f"fake query {ix}",
            "response": f"fake response {ix}",
            "context": f"fake context {ix}",
        }
        for ix in range(2)
    ]


@patch.object(RAGSystem, "retrieve")
def test_lsr_collator_with_mocks(
    mock_retrieve: MagicMock,
    mock_rag_system: RAGSystem,
    mock_examples: Sequence[dict],
    monkeypatch: MonkeyPatch,
) -> None:
    # Set environment variable for the duration of this test only
    monkeypatch.setenv("FEDRAG_SKIP_VALIDATION", "1")

    rag_system = RAGSystem(
        generator=mock_rag_system.generator,
        retriever=mock_rag_system.retriever,
        knowledge_store=mock_rag_system.knowledge_store,
        rag_config=mock_rag_system.rag_config,
    )

    # arrange mocks
    mock_tokenizer = MagicMock()
    mock_encode_return: EncodeResult = {
        "attention_mask": [1, 1, 1],
        "input_ids": [1, 2, 3],
    }
    mock_tokenizer.encode.return_value = mock_encode_return
    mock_tokenizer.unwrapped.pad_token_id = 42
    mock_tokenizer.unwrapped.pad_token = "<PAD>"
    rag_system.generator.tokenizer = mock_tokenizer

    # mock top-k = 2
    mock_top_k_val = 2
    mock_retrieve.side_effect = [
        [
            SourceNode(
                score=ix / 10,
                node=KnowledgeNode(
                    embedding=[ix, ix, ix],
                    node_type="text",
                    text_content=f"fake text context {ix}",
                ),
            )
            for ix in range(mock_top_k_val)
        ],
        [
            SourceNode(
                score=ix / 10,
                node=KnowledgeNode(
                    embedding=[ix, ix, ix],
                    node_type="text",
                    text_content=f"fake text context {ix}",
                ),
            )
            for ix in range(mock_top_k_val, mock_top_k_val * 2)
        ],
    ]

    # arrange collator
    collator = DataCollatorForRALT(
        rag_system=mock_rag_system,
        example_template="{query} {context} {response}",
    )

    # act
    collated_batch = collator(mock_examples)

    expected_num_examples = mock_top_k_val * len(mock_examples)
    assert collated_batch["input_ids"].shape[0] == expected_num_examples
    assert_close(
        collated_batch["input_ids"],
        torch.tensor([[1, 2, 3] for _ in range(expected_num_examples)]),
    )
    assert_close(
        collated_batch["labels"],
        torch.tensor([[1, 2, 3] for _ in range(expected_num_examples)]),
    )


def test_lsr_collator_raises_data_collator_error_when_pad_token_id_less_than_0(
    mock_rag_system: RAGSystem,
    monkeypatch: MonkeyPatch,
) -> None:
    monkeypatch.setenv("FEDRAG_SKIP_VALIDATION", "1")

    # arrange collator
    collator = DataCollatorForRALT(
        rag_system=mock_rag_system,
    )

    # arrange mock tokenizer
    mock_tokenizer = MagicMock()
    mock_tokenizer.pad_token = "<PAD>"
    mock_tokenizer.pad_token_id = -1

    # act
    with pytest.raises(
        DataCollatorError,
        match="Asking to pad but the tokenizer has a value for pad_token_id < 0.",
    ):
        collator._apply_padding(
            max_length=10,
            inputs_list=[[1, 1, 1]],
            attention_mask_list=[[2, 2, 2]],
            tokenizer=mock_tokenizer,
        )


def test_lsr_collator_raises_data_collator_error_when_pad_token_id_an(
    mock_rag_system: RAGSystem,
    monkeypatch: MonkeyPatch,
) -> None:
    monkeypatch.setenv("FEDRAG_SKIP_VALIDATION", "1")

    # arrange collator
    collator = DataCollatorForRALT(
        rag_system=mock_rag_system,
    )

    # arrange mock tokenizer
    mock_tokenizer = MagicMock()
    mock_tokenizer.pad_token = None
    mock_tokenizer.pad_token_id = 1
    mock_tokenizer.eos_token_id = None

    # act
    msg = (
        "Asking to pad but the tokenizer does not have a padding token "
        "nor an eos token that can potentially be used in its place."
    )
    with pytest.raises(
        DataCollatorError,
        match=msg,
    ):
        collator._apply_padding(
            max_length=10,
            inputs_list=[[1, 1, 1]],
            attention_mask_list=[[2, 2, 2]],
            tokenizer=mock_tokenizer,
        )


def test_lsr_collator_apply_padding(
    mock_rag_system: RAGSystem,
    monkeypatch: MonkeyPatch,
) -> None:
    monkeypatch.setenv("FEDRAG_SKIP_VALIDATION", "1")

    # arrange collator
    collator = DataCollatorForRALT(
        rag_system=mock_rag_system,
    )

    # arrange mock tokenizer
    mock_tokenizer = MagicMock()
    mock_tokenizer.pad_token = "<PAD>"
    mock_tokenizer.pad_token_id = 42

    # act
    batch = collator._apply_padding(
        max_length=3,
        inputs_list=[[1, 1, 1], [2, 2]],
        attention_mask_list=[[1, 1, 1], [1, 1]],
        tokenizer=mock_tokenizer,
    )

    # assert
    expected = {
        "input_ids": torch.tensor([[1, 1, 1], [42, 2, 2]]),
        "attention_mask": torch.tensor([[1, 1, 1], [0, 1, 1]]),
        "labels": torch.tensor([[1, 1, 1], [-100, 2, 2]]),
    }
    for k, v in batch.items():
        assert_close(v, expected[k])


def test_lsr_collator_apply_padding_with_eos_token(
    mock_rag_system: RAGSystem,
    monkeypatch: MonkeyPatch,
) -> None:
    monkeypatch.setenv("FEDRAG_SKIP_VALIDATION", "1")

    # arrange collator
    collator = DataCollatorForRALT(
        rag_system=mock_rag_system,
    )

    # arrange mock tokenizer
    mock_tokenizer = MagicMock()
    mock_tokenizer.pad_token = None
    mock_tokenizer.eos_token_id = 42

    # act
    batch = collator._apply_padding(
        max_length=3,
        inputs_list=[[1, 1, 1], [2, 2]],
        attention_mask_list=[[1, 1, 1], [1, 1]],
        tokenizer=mock_tokenizer,
    )

    # assert
    expected = {
        "input_ids": torch.tensor([[1, 1, 1], [42, 2, 2]]),
        "attention_mask": torch.tensor([[1, 1, 1], [0, 1, 1]]),
        "labels": torch.tensor([[1, 1, 1], [-100, 2, 2]]),
    }
    for k, v in batch.items():
        assert_close(v, expected[k])



from typing import Any

import numpy as np
import pytest
import torch
from pydantic import PrivateAttr
from torch.utils.data import DataLoader, Dataset

from fed_rag.base.generator import BaseGenerator
from fed_rag.base.retriever import BaseRetriever
from fed_rag.base.tokenizer import BaseTokenizer
from fed_rag.knowledge_stores.in_memory import InMemoryKnowledgeStore
from fed_rag.types.rag_system import RAGConfig, RAGSystem


class _TestDataset(Dataset):
    def __init__(self, size: int) -> None:
        self.features = np.random.rand(size, 2)
        self.labels = np.random.choice(2, size=size)

    def __len__(self) -> int:
        return len(self.labels)

    def __getitem__(self, index: int) -> tuple[np.ndarray, Any]:
        return self.features[index], self.labels[index]


@pytest.fixture()
def train_dataloader() -> DataLoader:
    dataset = _TestDataset(size=10)
    return DataLoader(dataset, batch_size=2, shuffle=True)


@pytest.fixture()
def val_dataloader() -> DataLoader:
    dataset = _TestDataset(size=4)
    return DataLoader(dataset, batch_size=2, shuffle=True)


class MockRetriever(BaseRetriever):
    _encoder: torch.nn.Module = PrivateAttr(default=torch.nn.Linear(3, 3))

    def encode_context(self, context: str, **kwargs: Any) -> torch.Tensor:
        return self._encoder.forward(torch.ones(3))

    def encode_query(self, query: str, **kwargs: Any) -> torch.Tensor:
        return self._encoder.forward(torch.zeros(3))

    @property
    def encoder(self) -> torch.nn.Module:
        return self._encoder

    @encoder.setter
    def encoder(self, value: torch.nn.Module) -> None:
        self._encoder = value

    @property
    def query_encoder(self) -> torch.nn.Module | None:
        return None

    @property
    def context_encoder(self) -> torch.nn.Module | None:
        return None


@pytest.fixture
def mock_retriever() -> MockRetriever:
    return MockRetriever()


class MockDualRetriever(BaseRetriever):
    _query_encoder: torch.nn.Module = PrivateAttr(
        default=torch.nn.Linear(2, 1)
    )
    _context_encoder: torch.nn.Module = PrivateAttr(
        default=torch.nn.Linear(2, 1)
    )

    def encode_context(self, context: str, **kwargs: Any) -> torch.Tensor:
        return self._encoder.forward(torch.ones(2))

    def encode_query(self, query: str, **kwargs: Any) -> torch.Tensor:
        return self._encoder.forward(torch.zeros(2))

    @property
    def encoder(self) -> torch.nn.Module | None:
        return None

    @property
    def query_encoder(self) -> torch.nn.Module | None:
        return self._query_encoder

    @query_encoder.setter
    def query_encoder(self, value: torch.nn.Module) -> None:
        self._query_encoder = value

    @property
    def context_encoder(self) -> torch.nn.Module | None:
        return self._context_encoder

    @context_encoder.setter
    def context_encoder(self, value: torch.nn.Module) -> None:
        self._context_encoder = value


@pytest.fixture
def mock_dual_retriever() -> MockDualRetriever:
    return MockDualRetriever()


class MockTokenizer(BaseTokenizer):
    def encode(self, input: str, **kwargs: Any) -> list[int]:
        return [0, 1, 2]

    def decode(self, input_ids: list[int], **kwargs: Any) -> str:
        return "mock decoded sentence"

    @property
    def unwrapped(self) -> None:
        return None


@pytest.fixture()
def mock_tokenizer() -> BaseTokenizer:
    return MockTokenizer()


class MockGenerator(BaseGenerator):
    _model = torch.nn.Linear(2, 1)
    _tokenizer = MockTokenizer()

    def generate(self, query: str, context: str, **kwargs: Any) -> str:
        return f"mock output from '{query}' and '{context}'."

    def compute_target_sequence_proba(self, prompt: str, target: str) -> float:
        return 0.42

    @property
    def model(self) -> torch.nn.Module:
        return self._model

    @model.setter
    def model(self, value: torch.nn.Module) -> None:
        self._model = value

    @property
    def tokenizer(self) -> MockTokenizer:
        return self._tokenizer


@pytest.fixture
def mock_generator() -> BaseGenerator:
    return MockGenerator()


@pytest.fixture
def mock_rag_system(
    mock_generator: BaseGenerator,
    mock_retriever: BaseRetriever,
) -> RAGSystem:
    knowledge_store = InMemoryKnowledgeStore()
    rag_config = RAGConfig(
        top_k=2,
    )
    return RAGSystem(
        generator=mock_generator,
        retriever=mock_retriever,
        knowledge_store=knowledge_store,
        rag_config=rag_config,
    )


@pytest.fixture
def another_mock_rag_system(
    mock_generator: BaseGenerator,
    mock_retriever: BaseRetriever,
) -> RAGSystem:
    knowledge_store = InMemoryKnowledgeStore()
    rag_config = RAGConfig(
        top_k=1,
    )
    return RAGSystem(
        generator=mock_generator,
        retriever=mock_retriever,
        knowledge_store=knowledge_store,
        rag_config=rag_config,
    )



import importlib.util

from fed_rag.exceptions.common import MissingExtraError

from .hf_peft_model import HFPeftModelGenerator
from .hf_pretrained_model import HFPretrainedModelGenerator

# check if huggingface extra is installed
_has_huggingface = (importlib.util.find_spec("transformers") is not None) and (
    importlib.util.find_spec("peft") is not None
)
if not _has_huggingface:
    msg = (
        f"`{__name__}` requires `huggingface` extra to be installed."
        " To fix please run `pip install fed-rag[huggingface]`."
    )
    raise MissingExtraError(msg)

__all__ = ["HFPeftModelGenerator", "HFPretrainedModelGenerator"]



"""HuggingFace Trainer Mixin"""

from abc import ABC, abstractmethod
from typing import (
    TYPE_CHECKING,
    Any,
    Optional,
    Protocol,
    Union,
    runtime_checkable,
)

from pydantic import BaseModel, ConfigDict

from fed_rag.exceptions import MissingExtraError

try:
    from datasets import Dataset
    from peft import PeftModel
    from sentence_transformers import SentenceTransformer
    from transformers import PreTrainedModel, Trainer, TrainingArguments

    _has_huggingface = True
except ModuleNotFoundError:
    _has_huggingface = False


if TYPE_CHECKING:  # pragma: no cover
    from datasets import Dataset
    from transformers import TrainingArguments


# Define the protocol for runtime checking
@runtime_checkable
class HuggingFaceTrainerProtocol(Protocol):
    train_dataset: "Dataset"
    training_arguments: Optional["TrainingArguments"]

    def model(
        self,
    ) -> Union["SentenceTransformer", "PreTrainedModel", "PeftModel"]:
        pass  # pragma: no cover


class HuggingFaceTrainerMixin(BaseModel, ABC):
    """HuggingFace Trainer Mixin."""

    model_config = ConfigDict(
        arbitrary_types_allowed=True,
    )
    train_dataset: "Dataset"
    training_arguments: Optional["TrainingArguments"] = None

    def __init__(
        self,
        train_dataset: "Dataset",
        training_arguments: Optional["TrainingArguments"] = None,
        **kwargs: Any,
    ):
        if not _has_huggingface:
            msg = (
                f"`{self.__class__.__name__}` requires `huggingface` extra to be installed. "
                "To fix please run `pip install fed-rag[huggingface]`."
            )
            raise MissingExtraError(msg)
        super().__init__(
            train_dataset=train_dataset,
            training_arguments=training_arguments,
            **kwargs,
        )

    @property
    @abstractmethod
    def hf_trainer_obj(self) -> "Trainer":
        """A ~transformers.Trainer object."""



from .core import FedRAGError


class RAGTrainerManagerError(FedRAGError):
    """Base errors for all rag trainer manager relevant exceptions."""

    pass


class UnspecifiedRetrieverTrainer(RAGTrainerManagerError):
    pass


class UnspecifiedGeneratorTrainer(RAGTrainerManagerError):
    pass


class UnsupportedTrainerMode(RAGTrainerManagerError):
    pass


class InconsistentRAGSystems(RAGTrainerManagerError):
    pass



"""PyTorchFLTask Unit Tests"""

from typing import Any, Callable

import numpy as np
import pytest
import torch.nn as nn
from torch.utils.data import DataLoader, Dataset

from fed_rag.decorators import federate
from fed_rag.types import TestResult, TrainResult


class _TestDataset(Dataset):
    def __init__(self, size: int) -> None:
        self.features = np.random.rand(size, 2)
        self.labels = np.random.choice(2, size=size)

    def __len__(self) -> int:
        return len(self.labels)

    def __getitem__(self, index: int) -> tuple[np.ndarray, Any]:
        return self.features[index], self.labels[index]


@pytest.fixture()
def train_dataloader() -> DataLoader:
    dataset = _TestDataset(size=10)
    return DataLoader(dataset, batch_size=2, shuffle=True)


@pytest.fixture()
def val_dataloader() -> DataLoader:
    dataset = _TestDataset(size=4)
    return DataLoader(dataset, batch_size=2, shuffle=True)


@pytest.fixture()
def trainer() -> Callable:
    @federate.trainer.pytorch
    def fn(
        net: nn.Module,
        train_loader: DataLoader,
        val_loader: DataLoader,
    ) -> TrainResult:
        return TrainResult(loss=0.0)

    return fn  # type: ignore


@pytest.fixture()
def tester() -> Callable:
    @federate.tester.pytorch
    def fn(
        net: nn.Module,
        test_loader: DataLoader,
    ) -> TestResult:
        return TestResult(loss=0.0, metrics={})

    return fn  # type: ignore


@pytest.fixture()
def mismatch_tester() -> Callable:
    @federate.tester.pytorch
    def fn(
        mdl: nn.Module,  # mismatch in name here
        test_loader: DataLoader,
    ) -> TestResult:
        return TestResult(loss=0.0, metrics={})

    return fn  # type: ignore



import pytest

from fed_rag.base.trainer import BaseGeneratorTrainer, BaseRetrieverTrainer
from fed_rag.base.trainer_manager import RAGTrainMode
from fed_rag.exceptions import InconsistentRAGSystems, UnsupportedTrainerMode
from fed_rag.types.rag_system import RAGSystem

from .conftest import MockRAGTrainerManager


def test_init(
    generator_trainer: BaseGeneratorTrainer,
    retriever_trainer: BaseRetrieverTrainer,
) -> None:
    trainer = MockRAGTrainerManager(
        generator_trainer=generator_trainer,
        retriever_trainer=retriever_trainer,
        mode="generator",
    )

    assert trainer.retriever_trainer == retriever_trainer
    assert trainer.generator_trainer == generator_trainer
    assert trainer.mode == "generator"
    assert trainer.model == generator_trainer.model

    # update mode
    trainer.mode = "retriever"
    assert trainer.model == retriever_trainer.model


def test_invalid_mode_raises_error(
    generator_trainer: BaseGeneratorTrainer,
    retriever_trainer: BaseRetrieverTrainer,
) -> None:
    msg = (
        f"Unsupported RAG train mode: both. "
        f"Mode must be one of: {', '.join([m.value for m in RAGTrainMode])}"
    )
    with pytest.raises(UnsupportedTrainerMode, match=msg):
        MockRAGTrainerManager(
            generator_trainer=generator_trainer,
            retriever_trainer=retriever_trainer,
            mode="both",
        )


def test_inconsistent_rag_systems_raises_error(
    generator_trainer: BaseGeneratorTrainer,
    retriever_trainer: BaseRetrieverTrainer,
    another_mock_rag_system: RAGSystem,
) -> None:
    msg = (
        "Inconsistent RAG systems detected between retriever and generator trainers. "
        "Both trainers must use the same RAG system instance for consistent training."
    )
    with pytest.raises(InconsistentRAGSystems, match=msg):
        generator_trainer.rag_system = another_mock_rag_system
        MockRAGTrainerManager(
            generator_trainer=generator_trainer,
            retriever_trainer=retriever_trainer,
            mode="generator",
        )



"""HuggingFace FL Task.

NOTE: Using this module requires the `huggingface` extra to be installed.
"""

import warnings
from typing import (
    TYPE_CHECKING,
    Any,
    Callable,
    Dict,
    OrderedDict,
    TypeAlias,
    Union,
    cast,
)

import torch
from flwr.client import NumPyClient
from flwr.client.client import Client
from flwr.common import NDArrays, Scalar
from flwr.common.parameter import ndarrays_to_parameters
from flwr.server.client_manager import ClientManager, SimpleClientManager
from flwr.server.server import Server
from flwr.server.strategy import FedAvg, Strategy
from pydantic import BaseModel, ConfigDict, PrivateAttr

from fed_rag.exceptions.common import MissingExtraError

# check if huggingface extra was installed
try:
    from datasets import Dataset
    from peft import PeftModel
    from peft.utils import get_peft_model_state_dict, set_peft_model_state_dict
    from sentence_transformers import SentenceTransformer
    from transformers import PreTrainedModel

    _has_huggingface = True
except ModuleNotFoundError:
    _has_huggingface = False

if TYPE_CHECKING:  # pragma: no cover
    from sentence_transformers import SentenceTransformer
    from peft import PeftModel
    from transformers import PreTrainedModel

from typing_extensions import Self

from fed_rag.base.fl_task import BaseFLTask
from fed_rag.exceptions import (
    MissingRequiredNetParam,
    MissingTesterSpec,
    MissingTrainerSpec,
    NetTypeMismatch,
    UnequalNetParamWarning,
)
from fed_rag.inspectors.pytorch import (
    TesterSignatureSpec,
    TrainerSignatureSpec,
)
from fed_rag.types import TestResult, TrainResult

HFModelType: TypeAlias = Union[
    "SentenceTransformer", "PreTrainedModel", "PeftModel"
]
HuggingFaceFlowerServer: TypeAlias = Server


class BaseFLTaskBundle(BaseModel):
    model_config = ConfigDict(arbitrary_types_allowed=True)
    net: HFModelType
    train_dataset: Dataset
    val_dataset: Dataset
    trainer: Callable
    tester: Callable
    extra_test_kwargs: Any
    extra_train_kwargs: Any

    def __init__(self, **kwargs: Any):
        if not _has_huggingface:
            msg = (
                f"`{self.__class__.__name__}` requires `huggingface` extra to be installed. "
                "To fix please run `pip install fed-rag[huggingface]`."
            )
            raise MissingExtraError(msg)

        super().__init__(**kwargs)


def _get_weights(net: HFModelType) -> NDArrays:
    """Get weights from net.

    NOTE: SentenceTransformer and PreTrainedModel are subclasses of torch.nn.Module
    """
    if isinstance(net, PeftModel):
        state_dict = get_peft_model_state_dict(net)
        state_dict = cast(Dict[str, Any], state_dict)
    elif isinstance(net, SentenceTransformer) or isinstance(
        net, PreTrainedModel
    ):
        state_dict = net.state_dict()
    else:
        raise ValueError("Invalid `net` type.")
    return [val.cpu().numpy() for _, val in state_dict.items()]


class HuggingFaceFlowerClient(NumPyClient):
    def __init__(
        self,
        task_bundle: BaseFLTaskBundle,
    ) -> None:
        super().__init__()
        self.task_bundle = task_bundle

    def __getattr__(self, name: str) -> Any:
        if name in self.task_bundle.model_fields:
            return getattr(self.task_bundle, name)
        else:
            return super().__getattr__(name)

    def get_weights(self) -> NDArrays:
        return _get_weights(self.net)

    def set_weights(self, parameters: NDArrays) -> None:
        if isinstance(self.net, PeftModel):
            # get state dict
            state_dict = get_peft_model_state_dict(self.net)
            state_dict = cast(Dict[str, Any], state_dict)

            # update state dict with supplied parameters
            params_dict = zip(state_dict.keys(), parameters)
            state_dict = OrderedDict(
                {k: torch.tensor(v) for k, v in params_dict}
            )
            set_peft_model_state_dict(self.net, state_dict)
        else:  # SentenceTransformer | PreTrainedModel
            # get state dict
            state_dict = self.net.state_dict()

            # update state dict with supplied parameters
            params_dict = zip(state_dict.keys(), parameters)
            state_dict = OrderedDict(
                {k: torch.tensor(v) for k, v in params_dict}
            )
            self.net.load_state_dict(state_dict, strict=True)

    def fit(
        self, parameters: NDArrays, config: dict[str, Scalar]
    ) -> tuple[NDArrays, int, dict[str, Scalar]]:
        self.set_weights(parameters)

        result: TrainResult = self.trainer(
            self.net,
            self.train_dataset,
            self.val_dataset,
            **self.task_bundle.extra_train_kwargs,
        )
        return (
            self.get_weights(),
            len(self.train_dataset),
            {"loss": result.loss},
        )

    def evaluate(
        self, parameters: NDArrays, config: dict[str, Scalar]
    ) -> tuple[float, int, dict[str, Scalar]]:
        self.set_weights(parameters)
        result: TestResult = self.tester(
            self.net, self.val_dataset, **self.extra_test_kwargs
        )
        return result.loss, len(self.val_dataset), result.metrics


class HuggingFaceFLTask(BaseFLTask):
    _client: NumPyClient = PrivateAttr()
    _trainer: Callable = PrivateAttr()
    _trainer_spec: TrainerSignatureSpec = PrivateAttr()
    _tester: Callable = PrivateAttr()
    _tester_spec: TesterSignatureSpec = PrivateAttr()

    def __init__(
        self,
        trainer: Callable,
        trainer_spec: TrainerSignatureSpec,
        tester: Callable,
        tester_spec: TesterSignatureSpec,
        **kwargs: Any,
    ) -> None:
        if not _has_huggingface:
            msg = (
                f"`{self.__class__.__name__}` requires `huggingface` extra to be installed. "
                "To fix please run `pip install fed-rag[huggingface]`."
            )
            raise MissingExtraError(msg)

        if (
            trainer_spec.net_parameter_class_name
            != tester_spec.net_parameter_class_name
        ):
            msg = (
                "`trainer`'s model class is not the same as that for `tester`."
            )
            raise NetTypeMismatch(msg)

        if trainer_spec.net_parameter != tester_spec.net_parameter:
            msg = (
                "`trainer`'s model parameter name is not the same as that for `tester`. "
                "Will use the name supplied in `trainer`."
            )
            warnings.warn(msg, UnequalNetParamWarning)

        super().__init__(**kwargs)
        self._trainer = trainer
        self._trainer_spec = trainer_spec
        self._tester = tester
        self._tester_spec = tester_spec

    @property
    def training_loop(self) -> Callable:
        return self._trainer

    @classmethod
    def from_trainer_and_tester(
        cls, trainer: Callable, tester: Callable
    ) -> Self:
        # extract trainer spec
        try:
            trainer_spec: TrainerSignatureSpec = getattr(
                trainer, "__fl_task_trainer_config"
            )
        except AttributeError:
            msg = "Cannot extract `TrainerSignatureSpec` from supplied `trainer`."
            raise MissingTrainerSpec(msg)

        # extract tester spec
        try:
            tester_spec: TesterSignatureSpec = getattr(
                tester, "__fl_task_tester_config"
            )
        except AttributeError:
            msg = (
                "Cannot extract `TesterSignatureSpec` from supplied `tester`."
            )
            raise MissingTesterSpec(msg)

        return cls(
            trainer=trainer,
            trainer_spec=trainer_spec,
            tester=tester,
            tester_spec=tester_spec,
        )

    def server(
        self,
        strategy: Strategy | None = None,
        client_manager: ClientManager | None = None,
        **kwargs: Any,
    ) -> HuggingFaceFlowerServer | None:
        if strategy is None:
            if self._trainer_spec.net_parameter not in kwargs:
                msg = f"Please pass in a model using the model param name {self._trainer_spec.net_parameter}."
                raise MissingRequiredNetParam(msg)

            model = kwargs.pop(self._trainer_spec.net_parameter)

            ndarrays = _get_weights(model)
            parameters = ndarrays_to_parameters(ndarrays)
            strategy = FedAvg(
                fraction_evaluate=1.0,
                initial_parameters=parameters,
            )

        if client_manager is None:
            client_manager = SimpleClientManager()

        return HuggingFaceFlowerServer(
            client_manager=client_manager, strategy=strategy
        )

    def client(self, **kwargs: Any) -> Client | None:
        # validate kwargs
        if self._trainer_spec.net_parameter not in kwargs:
            msg = f"Please pass in a model using the model param name {self._trainer_spec.net_parameter}."
            raise MissingRequiredNetParam(msg)
        # build bundle
        net = kwargs.pop(self._trainer_spec.net_parameter)
        train_dataset = kwargs.pop(self._trainer_spec.train_data_param)
        val_dataset = kwargs.pop(self._trainer_spec.val_data_param)

        bundle = BaseFLTaskBundle(
            net=net,
            train_dataset=train_dataset,
            val_dataset=val_dataset,
            extra_train_kwargs=kwargs,
            extra_test_kwargs={},  # TODO make this functional or get rid of it
            trainer=self._trainer,
            tester=self._tester,
        )
        return HuggingFaceFlowerClient(task_bundle=bundle)

    def simulate(self, num_clients: int, **kwargs: Any) -> Any:
        raise NotImplementedError

    @classmethod
    def from_configs(cls, trainer_cfg: Any, tester_cfg: Any) -> Any:
        raise NotImplementedError



"""PyTorch FL Task"""

import warnings
from typing import Any, Callable, OrderedDict, TypeAlias

import torch
import torch.nn as nn
from flwr.client import NumPyClient
from flwr.client.client import Client
from flwr.common import NDArrays, Scalar
from flwr.common.parameter import ndarrays_to_parameters
from flwr.server.client_manager import ClientManager, SimpleClientManager
from flwr.server.server import Server
from flwr.server.strategy import FedAvg, Strategy
from pydantic import BaseModel, ConfigDict, PrivateAttr
from torch.utils.data import DataLoader
from typing_extensions import Self

from fed_rag.base.fl_task import BaseFLTask, BaseFLTaskConfig
from fed_rag.exceptions import (
    MissingRequiredNetParam,
    MissingTesterSpec,
    MissingTrainerSpec,
    UnequalNetParamWarning,
)
from fed_rag.inspectors.pytorch import (
    TesterSignatureSpec,
    TrainerSignatureSpec,
)
from fed_rag.types import TestResult, TrainResult


class PyTorchFLTaskConfig(BaseFLTaskConfig):
    pass


class BaseFLTaskBundle(BaseModel):
    model_config = ConfigDict(arbitrary_types_allowed=True)
    net: nn.Module
    trainloader: DataLoader
    valloader: DataLoader
    trainer: Callable
    tester: Callable
    extra_test_kwargs: Any
    extra_train_kwargs: Any


def _get_weights(net: nn.Module) -> NDArrays:
    return [val.cpu().numpy() for _, val in net.state_dict().items()]


class PyTorchFlowerClient(NumPyClient):
    def __init__(
        self,
        task_bundle: BaseFLTaskBundle,
    ) -> None:
        super().__init__()
        self.task_bundle = task_bundle

    def __getattr__(self, name: str) -> Any:
        if name in self.task_bundle.model_fields:
            return getattr(self.task_bundle, name)
        else:
            return super().__getattr__(name)

    def get_weights(self) -> NDArrays:
        return _get_weights(self.net)

    def set_weights(self, parameters: NDArrays) -> None:
        params_dict = zip(self.net.state_dict().keys(), parameters)
        state_dict = OrderedDict({k: torch.tensor(v) for k, v in params_dict})
        self.net.load_state_dict(state_dict, strict=True)

    def fit(
        self, parameters: NDArrays, config: dict[str, Scalar]
    ) -> tuple[NDArrays, int, dict[str, Scalar]]:
        self.set_weights(parameters)

        result: TrainResult = self.trainer(
            self.net,
            self.trainloader,
            self.valloader,
            **self.task_bundle.extra_train_kwargs,
        )
        return (
            self.get_weights(),
            len(self.trainloader.dataset),
            {"loss": result.loss},
        )

    def evaluate(
        self, parameters: NDArrays, config: dict[str, Scalar]
    ) -> tuple[float, int, dict[str, Scalar]]:
        self.set_weights(parameters)
        result: TestResult = self.tester(
            self.net, self.valloader, **self.extra_test_kwargs
        )
        return result.loss, len(self.valloader.dataset), result.metrics


PyTorchFlowerServer: TypeAlias = Server


class PyTorchFLTask(BaseFLTask):
    _client: NumPyClient = PrivateAttr()
    _trainer: Callable = PrivateAttr()
    _trainer_spec: TrainerSignatureSpec = PrivateAttr()
    _tester: Callable = PrivateAttr()
    _tester_spec: TesterSignatureSpec = PrivateAttr()

    def __init__(
        self,
        trainer: Callable,
        trainer_spec: TrainerSignatureSpec,
        tester: Callable,
        tester_spec: TesterSignatureSpec,
        **kwargs: Any,
    ) -> None:
        if trainer_spec.net_parameter != tester_spec.net_parameter:
            msg = (
                "`trainer`'s model parameter name is not the same as that for `tester`. "
                "Will use the name supplied in `trainer`."
            )
            warnings.warn(msg, UnequalNetParamWarning)

        super().__init__(**kwargs)
        self._trainer = trainer
        self._trainer_spec = trainer_spec
        self._tester = tester
        self._tester_spec = tester_spec

    @property
    def training_loop(self) -> Callable:
        return self._trainer

    @classmethod
    def from_trainer_and_tester(
        cls, trainer: Callable, tester: Callable
    ) -> Self:
        # extract trainer spec
        try:
            trainer_spec: TrainerSignatureSpec = getattr(
                trainer, "__fl_task_trainer_config"
            )
        except AttributeError:
            msg = "Cannot extract `TrainerSignatureSpec` from supplied `trainer`."
            raise MissingTrainerSpec(msg)

        # extract tester spec
        try:
            tester_spec: TesterSignatureSpec = getattr(
                tester, "__fl_task_tester_config"
            )
        except AttributeError:
            msg = (
                "Cannot extract `TesterSignatureSpec` from supplied `tester`."
            )
            raise MissingTesterSpec(msg)

        return cls(
            trainer=trainer,
            trainer_spec=trainer_spec,
            tester=tester,
            tester_spec=tester_spec,
        )

    @classmethod
    def from_configs(cls, trainer_cfg: Any, tester_cfg: Any) -> Any:
        return super().from_configs(trainer_cfg, tester_cfg)

    def server(
        self,
        strategy: Strategy | None = None,
        client_manager: ClientManager | None = None,
        **kwargs: Any,
    ) -> PyTorchFlowerServer | None:
        if strategy is None:
            if self._trainer_spec.net_parameter not in kwargs:
                msg = f"Please pass in a model using the model param name {self._trainer_spec.net_parameter}."
                raise MissingRequiredNetParam(msg)

            model = kwargs.pop(self._trainer_spec.net_parameter)

            ndarrays = _get_weights(model)
            parameters = ndarrays_to_parameters(ndarrays)
            strategy = FedAvg(
                fraction_evaluate=1.0,
                initial_parameters=parameters,
            )

        if client_manager is None:
            client_manager = SimpleClientManager()

        return PyTorchFlowerServer(
            client_manager=client_manager, strategy=strategy
        )

    def client(self, **kwargs: Any) -> Client | None:
        # validate kwargs
        if self._trainer_spec.net_parameter not in kwargs:
            msg = f"Please pass in a model using the model param name {self._trainer_spec.net_parameter}."
            raise MissingRequiredNetParam(msg)
        # build bundle
        net = kwargs.pop(self._trainer_spec.net_parameter)
        trainloader = kwargs.pop(self._trainer_spec.train_data_param)
        valloader = kwargs.pop(self._trainer_spec.val_data_param)

        bundle = BaseFLTaskBundle(
            net=net,
            trainloader=trainloader,
            valloader=valloader,
            extra_train_kwargs=kwargs,
            extra_test_kwargs={},  # TODO make this functional or get rid of it
            trainer=self._trainer,
            tester=self._tester,
        )
        return PyTorchFlowerClient(task_bundle=bundle)

    def simulate(self, num_clients: int, **kwargs: Any) -> Any:
        raise NotImplementedError



"""Exceptions for loss."""

from .core import FedRAGError


class LossError(FedRAGError):
    """Base loss errors for all loss-related exceptions."""

    pass


class InvalidReductionParam(LossError):
    pass



from unittest.mock import MagicMock, patch

import pytest
import torch
from torch.testing import assert_close

from fed_rag.exceptions.loss import InvalidReductionParam
from fed_rag.loss.pytorch.lsr import LSRLoss, ReductionMode


def test_lsr_loss_init() -> None:
    loss = LSRLoss(reduction="sum")

    assert loss.reduction == ReductionMode.SUM


def test_invalid_reduction_raises_error() -> None:
    with pytest.raises(InvalidReductionParam):
        LSRLoss(reduction="invalid_reduction")


@pytest.mark.parametrize(
    ("reduction", "expected"), [("mean", 10.5), ("sum", 21)]
)
@patch("fed_rag.loss.pytorch.lsr.F")
def test_lsr_forward_with_mocks(
    mock_torch_functional: MagicMock, reduction: str, expected: float
) -> None:
    # arrange mocks
    mock_torch_functional.softmax.side_effect = iter(
        [torch.Tensor([1, 2, 3]), torch.Tensor([4, 5, 6])]
    )
    mock_torch_functional.kl_div.return_value = torch.Tensor(
        [[1, 2, 3], [4, 5, 6]]
    )

    loss = LSRLoss(reduction=reduction)
    retrieval_scores = torch.zeros(3)
    lm_logits = torch.zeros(3)
    out = loss(retrieval_scores, lm_logits)

    # assert
    mock_torch_functional.log_softmax.assert_called_once_with(
        retrieval_scores, dim=1
    )
    mock_torch_functional.softmax.assert_called_once_with(lm_logits, dim=1)
    mock_torch_functional.kl_div.assert_called_once()
    assert out == torch.Tensor([expected])


def test_lsr_expected_output_with_sample_data(
    retrieved_chunks: torch.Tensor,
    contexts: torch.Tensor,
    lm_scores: torch.Tensor,
) -> None:
    # retriever chunks probas
    retriever_scores = (retrieved_chunks * contexts).sum(dim=-1)
    retriever_scale = 1.0
    retriever_scores /= retriever_scale

    # lm chunks probas
    lm_scale = 1.0
    lm_scores /= lm_scale

    # lsr loss
    loss = LSRLoss(reduction="mean")
    out = loss(retriever_scores, lm_scores)

    assert retrieved_chunks.shape == (2, 3, 10)
    assert_close(out, torch.tensor(5.661825180053711))



"""Base Error Class for FedRAG."""


class FedRAGError(Exception):
    """Base error for all fed-rag exceptions."""

    pass


class FedRAGWarning(Warning):
    """Base warning for all fed-rag warnings."""



import re
from contextlib import nullcontext as does_not_raise
from typing import ClassVar
from unittest.mock import MagicMock, patch

import pytest
from pydantic import BaseModel

from fed_rag.base.bridge import BaseBridgeMixin, BridgeMetadata
from fed_rag.exceptions import (
    MissingExtraError,
    MissingSpecifiedConversionMethod,
)


class _TestBridgeMixin(BaseBridgeMixin):
    _bridge_version = "0.1.0"
    _bridge_extra = "my-bridge"
    _framework = "my-bridge-framework"
    _compatible_versions = ["0.1.x"]
    _method_name = "to_bridge"

    def to_bridge(self) -> None:
        self._validate_framework_installed()
        return None


# overwrite RAGSystem for this test
class RAGSystem(_TestBridgeMixin, BaseModel):
    bridges: ClassVar[dict[str, BridgeMetadata]] = {}

    @classmethod
    def _register_bridge(cls, metadata: BridgeMetadata) -> None:
        """To be used only by `BaseBridgeMixin`."""
        if metadata["framework"] not in cls.bridges:
            cls.bridges[metadata["framework"]] = metadata


def test_bridge_init() -> None:
    with does_not_raise():
        _TestBridgeMixin()


def test_bridge_get_metadata() -> None:
    bridge_mixin = (
        _TestBridgeMixin()
    )  # not really supposed to be instantiated on own

    metadata = bridge_mixin.get_bridge_metadata()

    assert metadata["bridge_version"] == "0.1.0"
    assert metadata["compatible_versions"] == ["0.1.x"]
    assert metadata["framework"] == "my-bridge-framework"
    assert metadata["method_name"] == "to_bridge"


def test_rag_system_registry() -> None:
    rag_system = RAGSystem()

    assert _TestBridgeMixin._framework in RAGSystem.bridges

    metadata = rag_system.bridges["my-bridge-framework"]

    assert metadata == _TestBridgeMixin.get_bridge_metadata()


@patch("fed_rag.base.bridge.importlib.util")
def test_validate_framework_installed(mock_importlib_util: MagicMock) -> None:
    mock_importlib_util.find_spec.return_value = None

    # with bridge-extra
    msg = (
        "`my-bridge-framework` module is missing but needs to be installed. "
        "To fix please run `pip install fed-rag[my-bridge]`."
    )
    with pytest.raises(MissingExtraError, match=re.escape(msg)):
        rag_system = RAGSystem()
        rag_system.to_bridge()

    # without bridge-extra
    msg = (
        "`my-bridge-framework` module is missing but needs to be installed. "
        "To fix please run `pip install my-bridge-framework`."
    )
    with pytest.raises(MissingExtraError, match=re.escape(msg)):
        RAGSystem._bridge_extra = None  # type:ignore [assignment]
        rag_system = RAGSystem()
        rag_system.to_bridge()


def test_invalid_mixin_raises_error() -> None:
    msg = "Bridge mixin for `mock` is missing conversion method `missing_method`."
    with pytest.raises(MissingSpecifiedConversionMethod, match=re.escape(msg)):

        class InvalidMixin(BaseBridgeMixin):
            _bridge_version = "0.1.0"
            _bridge_extra = None
            _framework = "mock"
            _compatible_versions = ["0.1.x"]
            _method_name = "missing_method"

        # overwrite RAGSystem for this test
        class RAGSystem(InvalidMixin, BaseModel):
            bridges: ClassVar[dict[str, BridgeMetadata]] = {}

            @classmethod
            def _register_bridge(cls, metadata: BridgeMetadata) -> None:
                """To be used only by `BaseBridgeMixin`."""
                if metadata["framework"] not in cls.bridges:
                    cls.bridges[metadata["framework"]] = metadata



"""Exceptions for Knowledge Stores."""

from .core import FedRAGError, FedRAGWarning


class KnowledgeStoreError(FedRAGError):
    """Base knowledge store error for all knowledge-store-related exceptions."""

    pass


class KnowledgeStoreWarning(FedRAGWarning):
    """Base knowledge store error for all knowledge-store-related warnings."""

    pass


class KnowledgeStoreNotFoundError(KnowledgeStoreError, FileNotFoundError):
    pass


class InvalidDistanceError(KnowledgeStoreError):
    pass


class LoadNodeError(KnowledgeStoreError):
    pass



import tempfile
from pathlib import Path
from unittest.mock import MagicMock, patch

import pytest

from fed_rag.base.knowledge_store import BaseKnowledgeStore
from fed_rag.exceptions import KnowledgeStoreNotFoundError
from fed_rag.knowledge_stores.in_memory import ManagedInMemoryKnowledgeStore
from fed_rag.types.knowledge_node import KnowledgeNode


@pytest.fixture
def text_nodes() -> list[KnowledgeNode]:
    return [
        KnowledgeNode(
            embedding=[1.0, 0.0, 1.0],
            node_type="text",
            text_content="node 1",
            metadata={"key1": "value1"},
        ),
        KnowledgeNode(
            embedding=[1.0, 0.0, 0.0],
            node_type="text",
            text_content="node 2",
            metadata={"key2": "value2"},
        ),
        KnowledgeNode(
            embedding=[1.0, 1.0, 0.0],
            node_type="text",
            text_content="node 3",
            metadata={"key3": "value3"},
        ),
    ]


def test_in_memory_knowledge_store_class() -> None:
    names_of_base_classes = [
        b.__name__ for b in ManagedInMemoryKnowledgeStore.__mro__
    ]
    assert BaseKnowledgeStore.__name__ in names_of_base_classes


def test_in_memory_knowledge_store_init() -> None:
    knowledge_store = ManagedInMemoryKnowledgeStore()

    assert knowledge_store.count == 0


def test_from_nodes(text_nodes: list[KnowledgeNode]) -> None:
    knowledge_store = ManagedInMemoryKnowledgeStore.from_nodes(
        nodes=text_nodes
    )

    assert knowledge_store.count == 3
    assert all(n.node_id in knowledge_store._data for n in text_nodes)


def test_delete_node(text_nodes: list[KnowledgeNode]) -> None:
    knowledge_store = ManagedInMemoryKnowledgeStore.from_nodes(
        nodes=text_nodes
    )

    assert knowledge_store.count == 3

    res = knowledge_store.delete_node(text_nodes[0].node_id)

    assert res is True
    assert knowledge_store.count == 2
    assert text_nodes[0].node_id not in knowledge_store._data


def test_delete_node_returns_false(text_nodes: list[KnowledgeNode]) -> None:
    knowledge_store = ManagedInMemoryKnowledgeStore.from_nodes(
        nodes=text_nodes
    )

    assert knowledge_store.count == 3

    res = knowledge_store.delete_node("non_included_id")

    assert res is False
    assert knowledge_store.count == 3
    assert all(n.node_id in knowledge_store._data for n in text_nodes)


def test_load_node(text_nodes: list[KnowledgeNode]) -> None:
    knowledge_store = ManagedInMemoryKnowledgeStore()
    assert knowledge_store.count == 0

    knowledge_store.load_node(text_nodes[-1])

    assert knowledge_store.count == 1
    assert text_nodes[-1].node_id in knowledge_store._data


def test_load_nodes(text_nodes: list[KnowledgeNode]) -> None:
    knowledge_store = ManagedInMemoryKnowledgeStore()
    assert knowledge_store.count == 0

    knowledge_store.load_nodes(text_nodes)

    assert knowledge_store.count == 3
    assert all(n.node_id in knowledge_store._data for n in text_nodes)


def test_clear(text_nodes: list[KnowledgeNode]) -> None:
    knowledge_store = ManagedInMemoryKnowledgeStore.from_nodes(
        nodes=text_nodes
    )
    assert knowledge_store.count == 3

    knowledge_store.clear()

    assert knowledge_store.count == 0
    assert all(n.node_id not in knowledge_store._data for n in text_nodes)


@pytest.mark.parametrize(
    ("query_emb", "top_k", "expected_node_ix"),
    [([1.0, 1.0, 1.0], 2, [0, 2]), ([0.5, 0.0, 0.0], 1, [1])],
    ids=[str([1.0, 1.0, 1.0]), str([0.5, 0.0, 0.0])],
)
def test_retrieve(
    query_emb: list[float],
    top_k: int,
    expected_node_ix: list[int],
    text_nodes: list[KnowledgeNode],
) -> None:
    # arrange
    knowledge_store = ManagedInMemoryKnowledgeStore.from_nodes(
        nodes=text_nodes
    )

    # act
    res = knowledge_store.retrieve(query_emb, top_k=top_k)

    # assert
    assert [el[1] for el in res] == [text_nodes[ix] for ix in expected_node_ix]


def test_persist(text_nodes: list[KnowledgeNode]) -> None:
    with tempfile.TemporaryDirectory() as dirpath:
        knowledge_store = ManagedInMemoryKnowledgeStore.from_nodes(
            nodes=text_nodes
        )
        knowledge_store.cache_dir = dirpath
        knowledge_store.persist()

        filename = (
            Path(dirpath)
            / knowledge_store.name
            / f"{knowledge_store.ks_id}.parquet"
        )
        assert filename.exists()


@patch("fed_rag.knowledge_stores.mixins.uuid")
def test_load(mock_uuid: MagicMock, text_nodes: list[KnowledgeNode]) -> None:
    mock_uuid.uuid4.return_value = "test_ks_id"
    with tempfile.TemporaryDirectory() as dirpath:
        knowledge_store = ManagedInMemoryKnowledgeStore.from_nodes(
            nodes=text_nodes, name="test_ks"
        )
        knowledge_store.cache_dir = dirpath
        knowledge_store.persist()

        loaded_knowledge_store = (
            ManagedInMemoryKnowledgeStore.from_name_and_id(
                name="test_ks", ks_id="test_ks_id", cache_dir=dirpath
            )
        )

        assert loaded_knowledge_store.ks_id == knowledge_store.ks_id
        assert loaded_knowledge_store._data == knowledge_store._data


def test_load_with_missing_file_raises_error() -> None:
    with tempfile.TemporaryDirectory() as dirpath:
        with pytest.raises(KnowledgeStoreNotFoundError):
            ManagedInMemoryKnowledgeStore.from_name_and_id(
                name="test_ks", ks_id="test_ks_id", cache_dir=dirpath
            )


@patch("fed_rag.knowledge_stores.mixins.uuid")
def test_persist_overwrite(
    mock_uuid: MagicMock,
    text_nodes: list[KnowledgeNode],
) -> None:
    mock_uuid.uuid4.return_value = "test_ks_id"
    with tempfile.TemporaryDirectory() as dirpath:
        knowledge_store = ManagedInMemoryKnowledgeStore.from_nodes(
            nodes=text_nodes, name="test_ks"
        )
        knowledge_store.cache_dir = dirpath
        knowledge_store.persist()

        knowledge_store.load_node(
            KnowledgeNode(
                embedding=[1.0, 1.0, 1.0],
                node_type="text",
                text_content="node 4",
                metadata={"key4": "value4"},
            )
        )
        knowledge_store.persist()

        loaded_knowledge_store = (
            ManagedInMemoryKnowledgeStore.from_name_and_id(
                name="test_ks", ks_id="test_ks_id", cache_dir=dirpath
            )
        )
        assert loaded_knowledge_store._data == knowledge_store._data



"""RAG System"""

from typing import Any, ClassVar

from pydantic import BaseModel, ConfigDict

from fed_rag.base.bridge import BridgeMetadata
from fed_rag.base.generator import BaseGenerator
from fed_rag.base.knowledge_store import BaseKnowledgeStore
from fed_rag.base.retriever import BaseRetriever
from fed_rag.bridges.llamaindex.bridge import LlamaIndexBridgeMixin
from fed_rag.types.knowledge_node import KnowledgeNode


class SourceNode(BaseModel):
    score: float
    node: KnowledgeNode

    def __getattr__(self, __name: str) -> Any:
        """Convenient wrapper on getattr of associated node."""
        return getattr(self.node, __name)


class RAGResponse(BaseModel):
    response: str
    source_nodes: list[SourceNode]

    def __str__(self) -> str:
        return self.response


class RAGConfig(BaseModel):
    top_k: int
    context_separator: str = "\n"


class RAGSystem(LlamaIndexBridgeMixin, BaseModel):
    model_config = ConfigDict(arbitrary_types_allowed=True)
    generator: BaseGenerator
    retriever: BaseRetriever
    knowledge_store: BaseKnowledgeStore
    rag_config: RAGConfig
    bridges: ClassVar[dict[str, BridgeMetadata]] = {}

    @classmethod
    def _register_bridge(cls, metadata: BridgeMetadata) -> None:
        """To be used only by `BaseBridgeMixin`."""
        if metadata["framework"] not in cls.bridges:
            cls.bridges[metadata["framework"]] = metadata

    def query(self, query: str) -> RAGResponse:
        """Query the RAG system."""
        source_nodes = self.retrieve(query)
        context = self._format_context(source_nodes)
        response = self.generate(query=query, context=context)
        return RAGResponse(source_nodes=source_nodes, response=response)

    def retrieve(self, query: str) -> list[SourceNode]:
        """Retrieve from KnowledgeStore."""
        query_emb: list[float] = self.retriever.encode_query(query).tolist()
        raw_retrieval_result = self.knowledge_store.retrieve(
            query_emb=query_emb, top_k=self.rag_config.top_k
        )
        return [
            SourceNode(score=el[0], node=el[1]) for el in raw_retrieval_result
        ]

    def generate(self, query: str, context: str) -> str:
        """Generate response to query with context."""
        return self.generator.generate(query=query, context=context)  # type: ignore

    def _format_context(self, source_nodes: list[SourceNode]) -> str:
        """Format the context from the source nodes."""
        # TODO: how to format image context
        return self.rag_config.context_separator.join(
            [node.get_content()["text_content"] for node in source_nodes]
        )



"""Decorators unit tests"""

from typing import Any

import pytest
import torch.nn as nn
from torch.utils.data import DataLoader

from fed_rag.decorators import federate
from fed_rag.exceptions.inspectors import (
    InvalidReturnType,
    MissingDataParam,
    MissingMultipleDataParams,
    MissingNetParam,
)
from fed_rag.inspectors.pytorch import (
    TesterSignatureSpec,
    TrainerSignatureSpec,
)
from fed_rag.types import TestResult, TrainResult


def test_decorated_trainer() -> None:
    def fn(
        net: nn.Module,
        train_loader: DataLoader,
        val_loader: DataLoader,
        extra_param_1: int,
        extra_param_2: float | None,
    ) -> TrainResult:
        pass

    decorated = federate.trainer.pytorch(fn)
    config: TrainerSignatureSpec = getattr(
        decorated, "__fl_task_trainer_config"
    )
    assert config.net_parameter == "net"
    assert config.train_data_param == "train_loader"
    assert config.val_data_param == "val_loader"
    assert config.extra_train_kwargs == ["extra_param_1", "extra_param_2"]
    assert config.net_parameter_class_name == "Module"


def test_decorated_trainer_raises_invalid_return_type_error() -> None:
    def fn(
        net: nn.Module,
        train_loader: DataLoader,
        val_loader: DataLoader,
        extra_param_1: int,
        extra_param_2: float | None,
    ) -> Any:
        pass

    with pytest.raises(
        InvalidReturnType,
        match="Trainer should return a fed_rag.types.TrainResult or a subclsas of it.",
    ):
        federate.trainer.pytorch(fn)


def test_decorated_trainer_raises_missing_net_param_error() -> None:
    def fn(
        train_loader: DataLoader,
        val_loader: DataLoader,
    ) -> TrainResult:
        pass

    with pytest.raises(MissingNetParam):
        federate.trainer.pytorch(fn)


def test_decorated_trainer_fails_to_find_two_data_params() -> None:
    def fn(
        model: nn.Module,
    ) -> TrainResult:
        pass

    msg = (
        "Inspection failed to find two data params for train and val datasets."
        "For PyTorch these params must be of type `torch.utils.data.DataLoader`"
    )
    with pytest.raises(MissingMultipleDataParams, match=msg):
        federate.trainer.pytorch(fn)


def test_decorated_trainer_fails_due_to_missing_data_loader() -> None:
    def fn(model: nn.Module, val_loader: DataLoader) -> TrainResult:
        pass

    msg = (
        "Inspection found one data param but failed to find another. "
        "Two data params are required for train and val datasets."
        "For PyTorch these params must be of type `torch.utils.data.DataLoader`"
    )
    with pytest.raises(MissingDataParam, match=msg):
        federate.trainer.pytorch(fn)


def test_decorated_trainer_from_instance_method() -> None:
    class _TestClass:
        @federate.trainer.pytorch
        def fn(
            self,
            net: nn.Module,
            train_loader: DataLoader,
            val_loader: DataLoader,
            extra_param_1: int,
            extra_param_2: float | None,
        ) -> TrainResult:
            pass

    obj = _TestClass()
    config: TrainerSignatureSpec = getattr(obj.fn, "__fl_task_trainer_config")
    assert config.net_parameter == "net"
    assert config.train_data_param == "train_loader"
    assert config.val_data_param == "val_loader"
    assert config.extra_train_kwargs == ["extra_param_1", "extra_param_2"]
    assert config.net_parameter_class_name == "Module"


def test_decorated_trainer_from_class_method() -> None:
    class _TestClass:
        @classmethod
        @federate.trainer.pytorch
        def fn(
            cls,
            net: nn.Module,
            train_loader: DataLoader,
            val_loader: DataLoader,
            extra_param_1: int,
            extra_param_2: float | None,
        ) -> TrainResult:
            pass

    obj = _TestClass()
    config: TrainerSignatureSpec = getattr(obj.fn, "__fl_task_trainer_config")
    assert config.net_parameter == "net"
    assert config.train_data_param == "train_loader"
    assert config.val_data_param == "val_loader"
    assert config.extra_train_kwargs == ["extra_param_1", "extra_param_2"]
    assert config.net_parameter_class_name == "Module"


def test_decorated_tester() -> None:
    def fn(
        mdl: nn.Module,
        test_loader: DataLoader,
        extra_param_1: int,
        extra_param_2: float | None,
    ) -> TestResult:
        pass

    decorated = federate.tester.pytorch(fn)
    config: TesterSignatureSpec = getattr(decorated, "__fl_task_tester_config")
    assert config.net_parameter == "mdl"
    assert config.test_data_param == "test_loader"
    assert config.extra_test_kwargs == ["extra_param_1", "extra_param_2"]
    assert config.net_parameter_class_name == "Module"


def test_decorated_tester_raises_invalid_return_type() -> None:
    def fn(
        mdl: nn.Module,
        test_loader: DataLoader,
        extra_param_1: int,
        extra_param_2: float | None,
    ) -> Any:
        pass

    with pytest.raises(
        InvalidReturnType,
        match="Tester should return a fed_rag.types.TestResult or a subclsas of it.",
    ):
        federate.tester.pytorch(fn)


def test_decorated_tester_raises_missing_net_param_error() -> None:
    def fn(
        test_loader: DataLoader,
    ) -> TestResult:
        pass

    with pytest.raises(MissingNetParam):
        federate.tester.pytorch(fn)


def test_decorated_tester_fails_to_a_data_params() -> None:
    def fn(
        model: nn.Module,
    ) -> TestResult:
        pass

    msg = (
        "Inspection failed to find a data param for a test dataset."
        "For PyTorch this params must be of type `torch.utils.data.DataLoader`"
    )
    with pytest.raises(MissingDataParam, match=msg):
        federate.tester.pytorch(fn)


def test_decorated_tester_from_instance_method() -> None:
    class _TestClass:
        @federate.tester.pytorch
        def fn(
            self,
            mdl: nn.Module,
            test_loader: DataLoader,
            extra_param_1: int,
        ) -> TestResult:
            pass

    obj = _TestClass()
    config: TesterSignatureSpec = getattr(obj.fn, "__fl_task_tester_config")
    assert config.net_parameter == "mdl"
    assert config.test_data_param == "test_loader"
    assert config.extra_test_kwargs == ["extra_param_1"]
    assert config.net_parameter_class_name == "Module"


def test_decorated_tester_from_class_method() -> None:
    class _TestClass:
        @classmethod
        @federate.tester.pytorch
        def fn(
            cls,
            mdl: nn.Module,
            test_loader: DataLoader,
            extra_param_1: int,
        ) -> TestResult:
            pass

    config: TesterSignatureSpec = getattr(
        _TestClass.fn, "__fl_task_tester_config"
    )
    assert config.net_parameter == "mdl"
    assert config.test_data_param == "test_loader"
    assert config.extra_test_kwargs == ["extra_param_1"]
    assert config.net_parameter_class_name == "Module"



"""PyTorch RAG Trainer"""

from typing import Any, Callable

import torch.nn as nn
from pydantic import BaseModel, Field
from torch.utils.data import DataLoader
from typing_extensions import assert_never

from fed_rag.base.trainer_manager import BaseRAGTrainerManager
from fed_rag.decorators import federate
from fed_rag.exceptions.trainer_manager import (
    UnspecifiedGeneratorTrainer,
    UnspecifiedRetrieverTrainer,
)
from fed_rag.fl_tasks.pytorch import PyTorchFLTask
from fed_rag.types.results import TestResult, TrainResult


class TrainingArgs(BaseModel):
    """Arguments for training."""

    learning_rate: float | None = None
    batch_size: int | None = None
    num_epochs: int | None = None
    warmup_steps: int | None = None
    weight_decay: float | None = None
    custom_kwargs: dict[str, Any] = Field(default_factory=dict)


class PyTorchRAGTrainerManager(BaseRAGTrainerManager):
    """PyTorch native RAG Trainer Manager"""

    def _prepare_generator_for_training(self, **kwargs: Any) -> None:
        self.generator_trainer.model.train()

        # freeze retriever
        if self.retriever_trainer:
            self.retriever_trainer.model.eval()

    def _prepare_retriever_for_training(
        self, freeze_context_encoder: bool = True, **kwargs: Any
    ) -> None:
        self.retriever_trainer.model.train()

        # freeze generator
        if self.generator_trainer:
            self.generator_trainer.model.eval()

    def _train_retriever(self, **kwargs: Any) -> None:
        if self.retriever_trainer:
            self._prepare_retriever_for_training()
            self.retriever_trainer.train()
        else:
            raise UnspecifiedRetrieverTrainer(
                "Attempted to perform retriever trainer with an unspecified trainer."
            )

    def _train_generator(self, **kwargs: Any) -> None:
        if self.generator_trainer:
            self._prepare_generator_for_training()
            self.generator_trainer.train()
        else:
            raise UnspecifiedGeneratorTrainer(
                "Attempted to perform generator trainer with an unspecified trainer."
            )

    def train(self, **kwargs: Any) -> None:
        if self.mode == "retriever":
            self._train_retriever()
        elif self.mode == "generator":
            self._train_generator()
        else:
            assert_never(self.mode)  # pragma: no cover

    def _get_federated_trainer(self) -> tuple[Callable, nn.Module]:
        if self.mode == "retriever":
            if self.retriever_trainer is None:
                raise UnspecifiedRetrieverTrainer(
                    "Cannot federate an unspecified retriever trainer."
                )
            retriever_train_fn = self.retriever_trainer.train
            retriever_module = self.retriever_trainer.model

            # Create a standalone function for federation
            def train_wrapper(
                _mdl: nn.Module,
                _train_dataloader: DataLoader,
                _val_dataloader: DataLoader,
            ) -> TrainResult:
                _ = retriever_train_fn()
                return TrainResult(loss=0)

            return federate.trainer.pytorch(train_wrapper), retriever_module

        elif self.mode == "generator":
            if self.generator_trainer is None:
                raise UnspecifiedGeneratorTrainer(
                    "Cannot federate an unspecified generator trainer."
                )
            generator_train_fn = self.generator_trainer.train
            generator_module = self.generator_trainer.model

            # Create a standalone function for federation
            def train_wrapper(
                _mdl: nn.Module,
                _train_dataloader: DataLoader,
                _val_dataloader: DataLoader,
            ) -> TrainResult:
                _ = generator_train_fn()
                # TODO get loss from out
                return TrainResult(loss=0)

            return federate.trainer.pytorch(train_wrapper), generator_module
        else:
            assert_never(self.mode)  # pragma: no cover

    def get_federated_task(self) -> PyTorchFLTask:
        federated_trainer, _module = self._get_federated_trainer()

        # TODO: add logic for getting evaluator/tester and then federate it as well
        # federated_tester = self.get_federated_tester(tester_decorator)
        # For now, using a simple placeholder test function
        def test_fn(_mdl: nn.Module, _dataloader: DataLoader) -> TestResult:
            # Implement simple testing or return a placeholder
            return TestResult(loss=0.42, metrics={})  # pragma: no cover

        federated_tester = federate.tester.pytorch(test_fn)

        return PyTorchFLTask.from_trainer_and_tester(
            trainer=federated_trainer,
            tester=federated_tester,
        )



from fed_rag.base.tokenizer import BaseTokenizer


def test_base(mock_tokenizer: BaseTokenizer) -> None:
    input_ids = mock_tokenizer.encode("hello world!")
    decoded_str = mock_tokenizer.decode([1, 2, 3])

    assert input_ids == [0, 1, 2]
    assert decoded_str == "mock decoded sentence"
    assert mock_tokenizer.unwrapped is None



"""Exceptions for inspectors."""

from .core import FedRAGError, FedRAGWarning


class InspectorError(FedRAGError):
    """Base inspector error for all inspector-related exceptions."""

    pass


class InspectorWarning(FedRAGWarning):
    """Base inspector warning for all inspector-related warnings."""

    pass


class MissingNetParam(InspectorError):
    pass


class MissingMultipleDataParams(InspectorError):
    pass


class MissingDataParam(InspectorError):
    pass


class MissingTrainerSpec(InspectorError):
    pass


class MissingTesterSpec(InspectorError):
    pass


class UnequalNetParamWarning(InspectorWarning):
    pass


class InvalidReturnType(InspectorError):
    pass



"""HuggingFace PretrainedTokenizer"""

from typing import TYPE_CHECKING, Any, Optional

from pydantic import ConfigDict, Field, PrivateAttr

from fed_rag.base.tokenizer import BaseTokenizer, EncodeResult
from fed_rag.exceptions.common import MissingExtraError

try:
    from transformers import AutoTokenizer, PreTrainedTokenizer

    _has_huggingface = True
except ModuleNotFoundError:
    _has_huggingface = False


if TYPE_CHECKING:  # pragma: no cover
    from transformers import PreTrainedTokenizer


class HFPretrainedTokenizer(BaseTokenizer):
    model_config = ConfigDict(protected_namespaces=("pydantic_model_",))
    model_name: str = Field(
        description="Name of HuggingFace model. Used for loading the model from HF hub or local."
    )
    load_model_kwargs: dict = Field(
        description="Optional kwargs dict for loading models from HF. Defaults to None.",
        default_factory=dict,
    )
    _tokenizer: Optional["PreTrainedTokenizer"] = PrivateAttr(default=None)

    def __init__(
        self,
        model_name: str,
        load_model_kwargs: dict | None = None,
        load_model_at_init: bool = True,
    ):
        if not _has_huggingface:
            msg = (
                f"`{self.__class__.__name__}` requires `huggingface` extra to be installed. "
                "To fix please run `pip install fed-rag[huggingface]`."
            )
            raise MissingExtraError(msg)
        super().__init__(
            model_name=model_name,
            load_model_kwargs=load_model_kwargs if load_model_kwargs else {},
        )
        if load_model_at_init:
            self._tokenizer = self._load_model_from_hf()

    def _load_model_from_hf(self, **kwargs: Any) -> "PreTrainedTokenizer":
        load_kwargs = self.load_model_kwargs
        load_kwargs.update(kwargs)
        self.load_model_kwargs = load_kwargs
        return AutoTokenizer.from_pretrained(self.model_name, **load_kwargs)

    @property
    def unwrapped(self) -> "PreTrainedTokenizer":
        if self._tokenizer is None:
            # load HF Pretrained Tokenizer
            tokenizer = self._load_model_from_hf()
            self._tokenizer = tokenizer
        return self._tokenizer

    @unwrapped.setter
    def unwrapped(self, value: "PreTrainedTokenizer") -> None:
        self._tokenizer = value

    def encode(self, input: str, **kwargs: Any) -> EncodeResult:
        tokenizer_result = self.unwrapped(text=input, **kwargs)
        retval: EncodeResult = {
            "input_ids": tokenizer_result.get("input_ids"),
            "attention_mask": tokenizer_result.get("attention_mask", None),
        }
        return retval

    def decode(self, input_ids: list[int], **kwargs: Any) -> str:
        return self.unwrapped.decode(token_ids=input_ids, **kwargs)  # type: ignore[no-any-return]



"""Base RAG Trainer Manager"""

from abc import ABC, abstractmethod
from enum import Enum
from typing import Any, cast

from pydantic import BaseModel, ConfigDict, field_validator, model_validator
from typing_extensions import assert_never

from fed_rag.base.trainer import BaseGeneratorTrainer, BaseRetrieverTrainer
from fed_rag.exceptions import (
    InconsistentRAGSystems,
    UnspecifiedGeneratorTrainer,
    UnspecifiedRetrieverTrainer,
    UnsupportedTrainerMode,
)

from .fl_task import BaseFLTask


class RAGTrainMode(str, Enum):
    RETRIEVER = "retriever"
    GENERATOR = "generator"


class BaseRAGTrainerManager(BaseModel, ABC):
    """Base RAG Trainer Class.

    The manager becomes solely responsible for orchestration, not for maintaining state
    (i.e., the RAGSystem).
    """

    model_config = ConfigDict(arbitrary_types_allowed=True)
    mode: RAGTrainMode
    retriever_trainer: BaseRetrieverTrainer | None = None
    generator_trainer: BaseGeneratorTrainer | None = None

    @field_validator("mode", mode="before")
    @classmethod
    def validate_mode(cls, v: str) -> str:
        try:
            # Try to convert to enum
            mode = RAGTrainMode(v)
            return mode
        except ValueError:
            # Catch the ValueError from enum conversion and raise your custom error
            raise UnsupportedTrainerMode(
                f"Unsupported RAG train mode: {v}. "
                f"Mode must be one of: {', '.join([m.value for m in RAGTrainMode])}"
            )

    # Validate trainer presence
    @model_validator(mode="after")
    def validate_trainers(self) -> "BaseRAGTrainerManager":
        """Validate trainer requirements."""
        # Validate trainer presence based on mode
        if (
            self.mode == RAGTrainMode.RETRIEVER
            and self.retriever_trainer is None
        ):
            raise UnspecifiedRetrieverTrainer(
                "Retriever trainer must be set when in retriever mode"
            )
        if (
            self.mode == RAGTrainMode.GENERATOR
            and self.generator_trainer is None
        ):
            raise UnspecifiedGeneratorTrainer(
                "Generator trainer must be set when in generator mode"
            )

        return self

    @model_validator(mode="after")
    def validate_trainers_consistency(self) -> "BaseRAGTrainerManager":
        """Validate that trainers use consistent RAG systems if both are present."""
        if (
            self.retriever_trainer is not None
            and self.generator_trainer is not None
        ):
            # Check if both trainers have the same RAG system reference
            if id(self.retriever_trainer.rag_system) != id(
                self.generator_trainer.rag_system
            ):
                raise InconsistentRAGSystems(
                    "Inconsistent RAG systems detected between retriever and generator trainers. "
                    "Both trainers must use the same RAG system instance for consistent training."
                )

        return self

    @abstractmethod
    def _prepare_retriever_for_training(
        self, freeze_context_encoder: bool = True, **kwargs: Any
    ) -> None:
        """Prepare retriever model for training."""

    @abstractmethod
    def _prepare_generator_for_training(self, **kwargs: Any) -> None:
        """Prepare generator model for training."""

    @abstractmethod
    def _train_retriever(self, **kwargs: Any) -> Any:
        """Train loop for retriever."""

    @abstractmethod
    def _train_generator(self, **kwargs: Any) -> Any:
        """Train loop for generator."""

    @abstractmethod
    def train(self, **kwargs: Any) -> Any:
        """Train loop for rag system."""

    @abstractmethod
    def get_federated_task(self) -> BaseFLTask:
        """Get the federated task."""

    @property
    def model(self) -> Any:
        """Return the model to be trained."""
        match self.mode:
            case RAGTrainMode.RETRIEVER:
                trainer = cast(BaseRetrieverTrainer, self.retriever_trainer)
            case RAGTrainMode.GENERATOR:
                trainer = cast(BaseGeneratorTrainer, self.generator_trainer)
            case _:  # pragma: no cover
                assert_never(self.mode)  # pragma: no cover
        return trainer.model



"""PyTorchFLTask Unit Tests"""

from typing import Callable, OrderedDict
from unittest.mock import MagicMock, patch

import pytest
import torch
from flwr.common.parameter import ndarrays_to_parameters
from flwr.server.client_manager import SimpleClientManager
from flwr.server.strategy import FedAvg
from torch.nn import Module
from torch.utils.data import DataLoader

from fed_rag.exceptions import MissingRequiredNetParam, UnequalNetParamWarning
from fed_rag.fl_tasks.pytorch import (
    BaseFLTaskBundle,
    PyTorchFlowerClient,
    PyTorchFLTask,
    _get_weights,
)
from fed_rag.types import TestResult, TrainResult


def test_init_flower_client(
    train_dataloader: DataLoader,
    val_dataloader: DataLoader,
    trainer: Callable,
    tester: Callable,
) -> None:
    bundle = BaseFLTaskBundle(
        net=torch.nn.Linear(2, 1),
        trainloader=train_dataloader,
        valloader=val_dataloader,
        trainer=trainer,
        tester=tester,
        extra_test_kwargs={},
        extra_train_kwargs={},
    )
    client = PyTorchFlowerClient(task_bundle=bundle)

    assert client.tester == tester
    assert client.trainer == trainer
    assert client.trainloader == train_dataloader
    assert client.valloader == val_dataloader
    assert client.extra_train_kwargs == {}
    assert client.extra_test_kwargs == {}


def test_flower_client_get_weights(
    train_dataloader: DataLoader,
    val_dataloader: DataLoader,
    trainer: Callable,
    tester: Callable,
) -> None:
    net = torch.nn.Linear(2, 1)
    bundle = BaseFLTaskBundle(
        net=net,
        trainloader=train_dataloader,
        valloader=val_dataloader,
        trainer=trainer,
        tester=tester,
        extra_test_kwargs={},
        extra_train_kwargs={},
    )
    client = PyTorchFlowerClient(task_bundle=bundle)
    expected_weights = [
        val.cpu().numpy() for _, val in net.state_dict().items()
    ]

    assert all((client.get_weights()[0] == expected_weights[0]).flatten())
    assert all((client.get_weights()[1] == expected_weights[1]).flatten())


@patch.object(Module, "load_state_dict")
def test_flower_client_set_weights(
    mock_load_state_dict: MagicMock,
    train_dataloader: DataLoader,
    val_dataloader: DataLoader,
    trainer: Callable,
    tester: Callable,
) -> None:
    net = torch.nn.Linear(2, 1)
    bundle = BaseFLTaskBundle(
        net=net,
        trainloader=train_dataloader,
        valloader=val_dataloader,
        trainer=trainer,
        tester=tester,
        extra_test_kwargs={},
        extra_train_kwargs={},
    )
    client = PyTorchFlowerClient(task_bundle=bundle)
    parameters = client.get_weights()

    # act
    client.set_weights(parameters)

    # assert
    mock_load_state_dict.assert_called_once()
    args, kwargs = mock_load_state_dict.call_args
    params_dict = zip(net.state_dict().keys(), parameters)
    state_dict = OrderedDict({k: torch.tensor(v) for k, v in params_dict})
    assert (args[0].get("weight") == state_dict["weight"]).all()
    assert kwargs == {"strict": True}


def test_flower_client_fit(
    train_dataloader: DataLoader,
    val_dataloader: DataLoader,
    trainer: Callable,
    tester: Callable,
) -> None:
    net = torch.nn.Linear(2, 1)
    bundle = BaseFLTaskBundle(
        net=net,
        trainloader=train_dataloader,
        valloader=val_dataloader,
        trainer=trainer,
        tester=tester,
        extra_test_kwargs={},
        extra_train_kwargs={},
    )
    client = PyTorchFlowerClient(task_bundle=bundle)
    parameters = client.get_weights()
    mock_trainer = MagicMock()
    client.trainer = mock_trainer
    mock_trainer.return_value = TrainResult(loss=0.01)

    # act
    result = client.fit(parameters, config={})

    # assert
    mock_trainer.assert_called_once()
    for a, b in zip(result[0], client.get_weights()):
        assert (a == b).all()
    assert result[1] == len(client.trainloader.dataset)
    assert result[2] == {"loss": 0.01}


def test_flower_client_evaluate(
    train_dataloader: DataLoader,
    val_dataloader: DataLoader,
    trainer: Callable,
    tester: Callable,
) -> None:
    net = torch.nn.Linear(2, 1)
    bundle = BaseFLTaskBundle(
        net=net,
        trainloader=train_dataloader,
        valloader=val_dataloader,
        trainer=trainer,
        tester=tester,
        extra_test_kwargs={},
        extra_train_kwargs={},
    )
    client = PyTorchFlowerClient(task_bundle=bundle)
    parameters = client.get_weights()
    mock_tester = MagicMock()
    client.tester = mock_tester
    mock_tester.return_value = TestResult(
        loss=0.01, metrics={"accuracy": 0.88}
    )

    # act
    result = client.evaluate(parameters=parameters, config={})

    # assert
    mock_tester.assert_called_once()
    assert result[0] == 0.01
    assert result[1] == len(client.valloader.dataset)
    assert result[2] == {"accuracy": 0.88}


def test_init_from_trainer_tester(trainer: Callable, tester: Callable) -> None:
    fl_task = PyTorchFLTask.from_trainer_and_tester(
        trainer=trainer, tester=tester
    )

    assert fl_task._trainer_spec == getattr(
        trainer, "__fl_task_trainer_config"
    )
    assert fl_task._tester_spec == getattr(tester, "__fl_task_tester_config")
    assert fl_task._tester == tester
    assert fl_task._trainer == trainer


def test_invoking_server_without_strategy_and_net_param_raises(
    trainer: Callable, tester: Callable
) -> None:
    fl_task = PyTorchFLTask.from_trainer_and_tester(
        trainer=trainer, tester=tester
    )
    with pytest.raises(
        MissingRequiredNetParam,
        match="Please pass in a model using the model param name net.",
    ):
        client_manager = SimpleClientManager()
        fl_task.server(client_manager=client_manager)


def test_invoking_server(trainer: Callable, tester: Callable) -> None:
    fl_task = PyTorchFLTask.from_trainer_and_tester(
        trainer=trainer, tester=tester
    )
    strategy = FedAvg()
    client_manager = SimpleClientManager()
    server = fl_task.server(strategy=strategy, client_manager=client_manager)

    assert server.client_manager() == client_manager
    assert server.strategy == strategy


def test_invoking_server_using_defaults(
    trainer: Callable, tester: Callable
) -> None:
    fl_task = PyTorchFLTask.from_trainer_and_tester(
        trainer=trainer, tester=tester
    )
    net = torch.nn.Linear(2, 1)
    ndarrays = _get_weights(net)
    parameters = ndarrays_to_parameters(ndarrays)

    # act
    server = fl_task.server(net=net)

    assert type(server.client_manager()) is SimpleClientManager
    assert type(server.strategy) is FedAvg
    assert server.strategy.initial_parameters == parameters


def test_invoking_client_without_net_param_raises(
    trainer: Callable, tester: Callable
) -> None:
    fl_task = PyTorchFLTask.from_trainer_and_tester(
        trainer=trainer, tester=tester
    )
    with pytest.raises(
        MissingRequiredNetParam,
        match="Please pass in a model using the model param name net.",
    ):
        fl_task.client()


def test_creating_fl_task_with_mismatched_net_params_raises_warning(
    trainer: Callable,
    mismatch_tester: Callable,
) -> None:
    msg = (
        "`trainer`'s model parameter name is not the same as that for `tester`. "
        "Will use the name supplied in `trainer`."
    )
    with pytest.warns(UnequalNetParamWarning, match=msg):
        PyTorchFLTask.from_trainer_and_tester(
            trainer=trainer,
            tester=mismatch_tester,
        )


def test_init_fl_task_with_mismatched_net_params_raises_warning(
    trainer: Callable,
    mismatch_tester: Callable,
) -> None:
    msg = (
        "`trainer`'s model parameter name is not the same as that for `tester`. "
        "Will use the name supplied in `trainer`."
    )
    with pytest.warns(UnequalNetParamWarning, match=msg):
        PyTorchFLTask(
            trainer=trainer,
            trainer_spec=trainer.__fl_task_trainer_config,  # type: ignore[attr-defined]
            tester=mismatch_tester,
            tester_spec=mismatch_tester.__fl_task_tester_config,  # type: ignore[attr-defined]
        )



from typing import Any, Optional, Sequence

from llama_index.core.base.base_query_engine import BaseQueryEngine
from llama_index.core.base.base_retriever import BaseRetriever
from llama_index.core.base.llms.types import (
    CompletionResponse,
    CompletionResponseGen,
    LLMMetadata,
)
from llama_index.core.data_structs.data_structs import IndexStruct
from llama_index.core.data_structs.struct_type import IndexStructType
from llama_index.core.indices.managed.base import BaseManagedIndex
from llama_index.core.llms.callbacks import llm_completion_callback
from llama_index.core.llms.custom import CustomLLM
from llama_index.core.llms.utils import LLMType
from llama_index.core.schema import Document, MediaResource
from llama_index.core.schema import Node as LlamaNode
from llama_index.core.schema import NodeWithScore, QueryBundle

from fed_rag.exceptions import BridgeError
from fed_rag.types.knowledge_node import KnowledgeNode
from fed_rag.types.rag_system import RAGSystem, SourceNode


def convert_source_node_to_llama_index_node_with_score(
    node: SourceNode,
) -> NodeWithScore:
    """Convert ~fed_rag.SourceNode to ~llama_index.NodeWithScore.

    NOTE: Currently only text nodes are supported.
    """
    text_resource = MediaResource(text=node.node.text_content)
    llama_index_node = LlamaNode(
        text_resource=text_resource, id_=node.node.node_id
    )
    return NodeWithScore(score=node.score, node=llama_index_node)


def convert_llama_index_node_to_knowledge_node(
    llama_node: LlamaNode,
) -> KnowledgeNode:
    """Convert ~llama_index.BaseNodes to ~fed_rag.KnowledgeNodes.

    NOTE: Currently only text nodes are supported.
    """
    if llama_node.embedding is None:
        raise BridgeError(
            "Failed to convert ~llama_index.Node: embedding attribute is None."
        )

    if llama_node.text_resource is None:
        raise BridgeError(
            "Failed to convert ~llama_index.Node: text_resource attribute is None."
        )
    return KnowledgeNode(
        node_id=llama_node.id_,
        embedding=llama_node.embedding,
        node_type="text",
        text_content=llama_node.text_resource.text,
        metadata=llama_node.metadata,
    )


class FedRAGManagedIndex(BaseManagedIndex):
    # Inner Clases
    class FedRAGRetriever(BaseRetriever):
        """A ~llama_index.BaseRetriever adapter for fed_rag.RAGSystem."""

        def __init__(self, rag_system: RAGSystem, *args: Any, **kwargs: Any):
            super().__init__(*args, **kwargs)
            self._rag_system = rag_system

        def _retrieve(self, query_bundle: QueryBundle) -> list[NodeWithScore]:
            """Retrieve specialization for FedRAG.

            Currently only supports text-based queries.
            """
            source_nodes = self._rag_system.retrieve(
                query=query_bundle.query_str
            )
            return [
                convert_source_node_to_llama_index_node_with_score(sn)
                for sn in source_nodes
            ]

    class FedRAGLLM(CustomLLM):
        """A ~llama_index.LLM adapter for fed_rag.RAGSystem.

        NOTE: this is a very basic LLM adapter.
        """

        def __init__(self, rag_system: RAGSystem, *args: Any, **kwargs: Any):
            super().__init__(*args, **kwargs)
            self._rag_system = rag_system

        @property
        def metadata(self) -> LLMMetadata:
            """Get LLM metadata."""
            return LLMMetadata(
                model_name="fedrag.generator",
            )

        @llm_completion_callback()
        def complete(self, prompt: str, **kwargs: Any) -> CompletionResponse:
            res = self._rag_system.generator.generate(query=prompt, context="")
            return CompletionResponse(text=res)

        @llm_completion_callback()
        def stream_complete(
            self, prompt: str, **kwargs: Any
        ) -> CompletionResponseGen:
            raise NotImplementedError(
                "stream_complete is not implemented for FedRAGLLM."
            )

    class FedRAGIndexStruct(IndexStruct):
        @classmethod
        def get_type(cls) -> IndexStructType:
            return IndexStructType.VECTOR_STORE

    # methods and attributes
    def __init__(self, rag_system: RAGSystem, *args: Any, **kwargs: Any):
        nodes = kwargs.get("nodes", [])
        if len(list(nodes)) > 0:
            raise BridgeError(
                "FedRAGManagedIndex does not support nodes on initialization."
            )
        super().__init__(nodes=[], *args, **kwargs)
        self._rag_system = rag_system

    def _insert(
        self, nodes: Sequence[LlamaNode], **insert_kwargs: Any
    ) -> None:
        knowledge_nodes = [
            convert_llama_index_node_to_knowledge_node(llama_node=n)
            for n in nodes
        ]
        self._rag_system.knowledge_store.load_nodes(knowledge_nodes)

    def delete_ref_doc(
        self,
        ref_doc_id: str,
        delete_from_docstore: bool = False,
        **delete_kwargs: Any,
    ) -> None:
        raise NotImplementedError(
            "_delete_ref_doc not implemented for `FedRAGManagedIndex`."
        )

    def _delete_node(self, node_id: str, **delete_kwargs: Any) -> None:
        # node id's are presereved after conversion
        self._rag_system.knowledge_store.delete_node(node_id=node_id)

    def update_ref_doc(self, document: Document, **update_kwargs: Any) -> None:
        raise NotImplementedError(
            "update_ref_doc not implemented for `FedRAGManagedIndex`."
        )

    def as_retriever(self, **kwargs: Any) -> BaseRetriever:
        return self.FedRAGRetriever(rag_system=self._rag_system)

    def as_query_engine(
        self, llm: Optional[LLMType] = None, **kwargs: Any
    ) -> BaseQueryEngine:
        # set llm
        llm = self.FedRAGLLM(rag_system=self._rag_system)
        return super().as_query_engine(llm=llm, **kwargs)

    def _build_index_from_nodes(
        self, nodes: Sequence[LlamaNode], **build_kwargs: Any
    ) -> IndexStruct:
        return self.FedRAGIndexStruct(
            summary="~fed_rag.FedRAGManagedIndex wrapper of RAGSystem."
        )



import re
import sys
from unittest.mock import MagicMock, patch

import pytest
from transformers import AutoTokenizer, PreTrainedTokenizer

from fed_rag.base.tokenizer import BaseTokenizer
from fed_rag.exceptions import MissingExtraError
from fed_rag.tokenizers.hf_pretrained_tokenizer import HFPretrainedTokenizer


def test_hf_pretrained_generator_class() -> None:
    names_of_base_classes = [b.__name__ for b in HFPretrainedTokenizer.__mro__]
    assert BaseTokenizer.__name__ in names_of_base_classes


@patch.object(HFPretrainedTokenizer, "_load_model_from_hf")
def test_hf_pretrained_tokenizer_class_init_delayed_load(
    mock_load_from_hf: MagicMock, hf_tokenizer: PreTrainedTokenizer
) -> None:
    tokenizer = HFPretrainedTokenizer(
        model_name="fake_name", load_model_at_init=False
    )

    assert tokenizer.model_name == "fake_name"
    assert tokenizer._tokenizer is None

    # load model
    mock_load_from_hf.return_value = hf_tokenizer

    tokenizer._load_model_from_hf()
    args, kwargs = mock_load_from_hf.call_args

    mock_load_from_hf.assert_called_once()
    assert tokenizer.unwrapped == hf_tokenizer
    assert args == ()
    assert kwargs == {}


@patch.object(HFPretrainedTokenizer, "_load_model_from_hf")
def test_hf_pretrained_generator_class_init(
    mock_load_from_hf: MagicMock, hf_tokenizer: PreTrainedTokenizer
) -> None:
    # arrange
    mock_load_from_hf.return_value = hf_tokenizer

    # act
    generator = HFPretrainedTokenizer(
        model_name="fake_name",
    )
    args, kwargs = mock_load_from_hf.call_args

    # assert
    mock_load_from_hf.assert_called_once()
    assert generator.model_name == "fake_name"
    assert generator.unwrapped == hf_tokenizer
    assert args == ()
    assert kwargs == {}


@patch.object(HFPretrainedTokenizer, "_load_model_from_hf")
def test_hf_pretrained_generator_class_init_no_load(
    mock_load_from_hf: MagicMock, hf_tokenizer: PreTrainedTokenizer
) -> None:
    tokenizer = HFPretrainedTokenizer(
        model_name="fake_name", load_model_at_init=False
    )

    mock_load_from_hf.assert_not_called()
    assert tokenizer.model_name == "fake_name"
    assert tokenizer._tokenizer is None

    # load model using setter
    tokenizer.unwrapped = hf_tokenizer

    assert tokenizer.unwrapped == hf_tokenizer


@patch.object(AutoTokenizer, "from_pretrained")
def test_hf_pretrained_load_model_from_hf(
    mock_tokenizer_from_pretrained: MagicMock,
    hf_tokenizer: PreTrainedTokenizer,
) -> None:
    # arrange
    mock_tokenizer_from_pretrained.return_value = hf_tokenizer

    # act
    tokenizer = HFPretrainedTokenizer(
        model_name="fake_name", load_model_kwargs={"device_map": "cpu"}
    )

    # assert
    assert tokenizer.model_name == "fake_name"
    mock_tokenizer_from_pretrained.assert_called_once_with(
        "fake_name", device_map="cpu"
    )
    assert tokenizer.unwrapped == hf_tokenizer


def test_encode() -> None:
    # arrange
    tokenizer = HFPretrainedTokenizer(
        model_name="fake_name", load_model_at_init=False
    )
    mock_tokenizer = MagicMock()
    mock_tokenizer.return_value = {
        "input_ids": [1, 2],
        "attention_mask": [1, 1],
    }
    tokenizer.unwrapped = mock_tokenizer

    # act
    result = tokenizer.encode("fake input")

    assert result["input_ids"] == [1, 2]
    assert result["attention_mask"] == [1, 1]
    mock_tokenizer.assert_called_once()


def test_decode() -> None:
    # arrange
    tokenizer = HFPretrainedTokenizer(
        model_name="fake_name", load_model_at_init=False
    )
    mock_tokenizer = MagicMock()
    mock_tokenizer.decode.return_value = "fake decoded"
    tokenizer.unwrapped = mock_tokenizer

    # act
    result = tokenizer.decode([1, 2])

    assert result == "fake decoded"
    mock_tokenizer.decode.assert_called_once_with(token_ids=[1, 2])


def test_huggingface_extra_missing() -> None:
    """Test extra is not installed."""

    modules = {"transformers": None}
    module_to_import = "fed_rag.tokenizers.hf_pretrained_tokenizer"

    if module_to_import in sys.modules:
        original_module = sys.modules.pop(module_to_import)

    with patch.dict("sys.modules", modules):
        msg = (
            "`HFPretrainedTokenizer` requires `huggingface` extra to be installed. "
            "To fix please run `pip install fed-rag[huggingface]`."
        )
        with pytest.raises(
            MissingExtraError,
            match=re.escape(msg),
        ):
            from fed_rag.tokenizers.hf_pretrained_tokenizer import (
                HFPretrainedTokenizer,
            )

            HFPretrainedTokenizer("fake_name")

    # restore module so to not affect other tests
    sys.modules[module_to_import] = original_module



"""Base Data Collator"""

from abc import ABC, abstractmethod
from typing import Any

from pydantic import BaseModel, ConfigDict

from fed_rag.types.rag_system import RAGSystem


class BaseDataCollator(BaseModel, ABC):
    """Base Data Collator."""

    model_config = ConfigDict(arbitrary_types_allowed=True)
    rag_system: RAGSystem

    @abstractmethod
    def __call__(self, features: list[dict[str, Any]], **kwargs: Any) -> Any:
        """Collate examples into a batch."""



"""Common abstractions for inspectors"""

from pydantic import BaseModel


class TrainerSignatureSpec(BaseModel):
    net_parameter: str
    train_data_param: str
    val_data_param: str
    extra_train_kwargs: list[str] = []
    net_parameter_class_name: str


class TesterSignatureSpec(BaseModel):
    __test__ = (
        False  # needed for Pytest collision. Avoids PytestCollectionWarning
    )
    net_parameter: str
    test_data_param: str
    extra_test_kwargs: list[str] = []
    net_parameter_class_name: str



import re
import sys
from contextlib import nullcontext as does_not_raise
from unittest.mock import MagicMock, patch

import pytest
from datasets import Dataset
from pytest import MonkeyPatch

from fed_rag.base.trainer import (
    BaseGeneratorTrainer,
    BaseRetrieverTrainer,
    BaseTrainer,
)
from fed_rag.base.trainer_manager import BaseRAGTrainerManager, RAGTrainMode
from fed_rag.exceptions import (
    MissingExtraError,
    UnspecifiedGeneratorTrainer,
    UnspecifiedRetrieverTrainer,
    UnsupportedTrainerMode,
)
from fed_rag.fl_tasks.huggingface import HuggingFaceFLTask
from fed_rag.trainer_managers.huggingface import HuggingFaceRAGTrainerManager


def test_pt_rag_trainer_class() -> None:
    names_of_base_classes = [
        b.__name__ for b in HuggingFaceRAGTrainerManager.__mro__
    ]
    assert BaseRAGTrainerManager.__name__ in names_of_base_classes


def test_init(
    retriever_trainer: BaseRetrieverTrainer,
    generator_trainer: BaseGeneratorTrainer,
    train_dataset: Dataset,
    monkeypatch: MonkeyPatch,
) -> None:
    monkeypatch.setenv("FEDRAG_SKIP_VALIDATION", "1")

    manager = HuggingFaceRAGTrainerManager(
        mode="retriever",
        train_dataset=train_dataset,
        retriever_trainer=retriever_trainer,
        generator_trainer=generator_trainer,
    )

    assert manager.generator_trainer == generator_trainer
    assert manager.retriever_trainer == retriever_trainer
    assert manager.mode == RAGTrainMode.RETRIEVER


def test_init_raises_unspecified_generator_trainer_error(
    monkeypatch: MonkeyPatch,
) -> None:
    # skip validation of rag system
    monkeypatch.setenv("FEDRAG_SKIP_VALIDATION", "1")

    with pytest.raises(
        UnspecifiedGeneratorTrainer,
        match="Generator trainer must be set when in generator mode",
    ):
        HuggingFaceRAGTrainerManager(
            mode="generator",
        )


@patch.object(HuggingFaceRAGTrainerManager, "_prepare_retriever_for_training")
def test_init_raises_unspecified_retriever_trainer_error(
    mock_prepare_retriever_for_training: MagicMock,
    monkeypatch: MonkeyPatch,
) -> None:
    # skip validation of rag system
    monkeypatch.setenv("FEDRAG_SKIP_VALIDATION", "1")

    with pytest.raises(
        UnspecifiedRetrieverTrainer,
        match="Retriever trainer must be set when in retriever mode",
    ):
        HuggingFaceRAGTrainerManager(
            mode="retriever",
        )


def test_huggingface_extra_missing(
    retriever_trainer: BaseRetrieverTrainer,
    generator_trainer: BaseGeneratorTrainer,
) -> None:
    modules = {
        "datasets": None,
    }
    module_to_import = "fed_rag.trainer_managers.huggingface"
    original_module = sys.modules.pop(module_to_import, None)

    with patch.dict("sys.modules", modules):
        msg = (
            "`HuggingFaceRAGTrainerManager` requires `huggingface` extra to be installed. "
            "To fix please run `pip install fed-rag[huggingface]`."
        )
        with pytest.raises(
            MissingExtraError,
            match=re.escape(msg),
        ):
            from fed_rag.trainer_managers.huggingface import (
                HuggingFaceRAGTrainerManager,
            )

            HuggingFaceRAGTrainerManager(
                mode="retriever",
                retriever_trainer=retriever_trainer,
                generator_trainer=generator_trainer,
            )

    # restore module so to not affect other tests
    if original_module:
        sys.modules[module_to_import] = original_module


@patch.object(HuggingFaceRAGTrainerManager, "_prepare_retriever_for_training")
def test_train_retriever(
    mock_prepare_retriever_for_training: MagicMock,
    retriever_trainer: BaseRetrieverTrainer,
    monkeypatch: MonkeyPatch,
) -> None:
    monkeypatch.setenv("FEDRAG_SKIP_VALIDATION", "1")

    mock_retriever_trainer = MagicMock()
    manager = HuggingFaceRAGTrainerManager(
        mode="retriever",
        retriever_trainer=retriever_trainer,
    )
    manager.retriever_trainer = mock_retriever_trainer

    manager.train()

    mock_prepare_retriever_for_training.assert_called_once()
    mock_retriever_trainer.train.assert_called_once_with()


@patch.object(HuggingFaceRAGTrainerManager, "_prepare_generator_for_training")
def test_train_generator(
    mock_prepare_generator_for_training: MagicMock,
    generator_trainer: BaseGeneratorTrainer,
    monkeypatch: MonkeyPatch,
) -> None:
    # skip validation of rag system
    monkeypatch.setenv("FEDRAG_SKIP_VALIDATION", "1")

    manager = HuggingFaceRAGTrainerManager(
        mode="generator",
        generator_trainer=generator_trainer,
    )
    mock_generator_trainer = MagicMock()
    manager.generator_trainer = mock_generator_trainer

    manager.train()

    mock_prepare_generator_for_training.assert_called_once()
    mock_generator_trainer.train.assert_called_once_with()


def test_get_federated_task_retriever(
    retriever_trainer: BaseTrainer,
    monkeypatch: MonkeyPatch,
) -> None:
    # skip validation of rag system
    monkeypatch.setenv("FEDRAG_SKIP_VALIDATION", "1")

    # arrange
    trainer = HuggingFaceRAGTrainerManager(
        mode="retriever",
        retriever_trainer=retriever_trainer,
    )

    # act
    retriever_trainer, _ = trainer._get_federated_trainer()
    out = retriever_trainer(MagicMock(), MagicMock(), MagicMock())
    fl_task = trainer.get_federated_task()

    # assert
    assert out.loss == 0
    assert isinstance(fl_task, HuggingFaceFLTask)
    assert fl_task._trainer_spec == retriever_trainer.__fl_task_trainer_config


def test_get_federated_task_generator(
    generator_trainer: BaseTrainer,
    monkeypatch: MonkeyPatch,
) -> None:
    # skip validation of rag system
    monkeypatch.setenv("FEDRAG_SKIP_VALIDATION", "1")

    # arrange
    trainer = HuggingFaceRAGTrainerManager(
        mode="generator",
        generator_trainer=generator_trainer,
    )

    # act
    generator_trainer, _ = trainer._get_federated_trainer()
    out = generator_trainer(MagicMock(), MagicMock(), MagicMock())
    fl_task = trainer.get_federated_task()

    # assert
    assert out.loss == 0
    assert isinstance(fl_task, HuggingFaceFLTask)
    assert fl_task._trainer_spec == generator_trainer.__fl_task_trainer_config


def test_invalid_mode_raises_error(
    monkeypatch: MonkeyPatch,
) -> None:
    # skip validation of rag system
    monkeypatch.setenv("FEDRAG_SKIP_VALIDATION", "1")

    msg = (
        f"Unsupported RAG train mode: both. "
        f"Mode must be one of: {', '.join([m.value for m in RAGTrainMode])}"
    )
    with pytest.raises(UnsupportedTrainerMode, match=msg):
        HuggingFaceRAGTrainerManager(
            mode="both",
        )


def test_get_federated_task_raises_unspecified_generator_error(
    monkeypatch: MonkeyPatch,
    retriever_trainer: BaseRetrieverTrainer,
) -> None:
    # skip validation of rag system
    monkeypatch.setenv("FEDRAG_SKIP_VALIDATION", "1")

    # arrange
    # this will pass validations at init
    manager = HuggingFaceRAGTrainerManager(
        mode="retriever", retriever_trainer=retriever_trainer
    )

    with pytest.raises(
        UnspecifiedGeneratorTrainer,
        match="Cannot federate an unspecified generator trainer.",
    ):
        manager.mode = "generator"  # user modifies the mode
        manager.get_federated_task()


def test_private_get_federated_task_raises_unspecified_retriever_error(
    monkeypatch: MonkeyPatch,
    generator_trainer: BaseGeneratorTrainer,
) -> None:
    # skip validation of rag system
    monkeypatch.setenv("FEDRAG_SKIP_VALIDATION", "1")

    # arrange
    # this will pass validations at init
    manager = HuggingFaceRAGTrainerManager(
        mode="generator", generator_trainer=generator_trainer
    )

    with pytest.raises(
        UnspecifiedRetrieverTrainer,
        match="Cannot federate an unspecified retriever trainer.",
    ):
        # change mode to retriever
        manager.mode = "retriever"
        manager.get_federated_task()


def test_prepare_generator_for_training(
    generator_trainer: BaseGeneratorTrainer,
    retriever_trainer: BaseRetrieverTrainer,
    monkeypatch: MonkeyPatch,
) -> None:
    monkeypatch.setenv("FEDRAG_SKIP_VALIDATION", "1")
    manager = HuggingFaceRAGTrainerManager(
        mode="generator",
        generator_trainer=generator_trainer,
        retriever_trainer=retriever_trainer,
    )

    # add mocks
    mock_generator_model = MagicMock()
    mock_retriever_model = MagicMock()
    manager.generator_trainer.model = mock_generator_model
    manager.retriever_trainer.model = mock_retriever_model

    # Check parameters before
    for param in generator_trainer.model.parameters():
        assert param.requires_grad is True  # They should start unfrozen

    # act
    with does_not_raise():
        manager._prepare_generator_for_training()

    # assert
    mock_generator_model.train.assert_called_once()
    mock_retriever_model.eval.assert_called_once()
    for param in generator_trainer.model.parameters():
        assert param.requires_grad is False  # They should now be frozen


def test_prepare_retriever_for_training(
    generator_trainer: BaseGeneratorTrainer,
    retriever_trainer: BaseRetrieverTrainer,
    monkeypatch: MonkeyPatch,
) -> None:
    monkeypatch.setenv("FEDRAG_SKIP_VALIDATION", "1")
    manager = HuggingFaceRAGTrainerManager(
        mode="retriever",
        generator_trainer=generator_trainer,
        retriever_trainer=retriever_trainer,
    )

    # add mocks
    mock_generator_model = MagicMock()
    mock_retriever_model = MagicMock()
    manager.generator_trainer.model = mock_generator_model
    manager.retriever_trainer.model = mock_retriever_model

    # Check parameters before
    for param in retriever_trainer.model.parameters():
        assert param.requires_grad is True  # They should start unfrozen

    # act
    with does_not_raise():
        manager._prepare_retriever_for_training()

    # assert
    mock_generator_model.eval.assert_called_once()
    mock_retriever_model.train.assert_called_once()
    for param in retriever_trainer.model.parameters():
        assert param.requires_grad is False  # They should now be frozen


def test_private_train_retriever_raises_unspecified_retriever_error(
    generator_trainer: BaseGeneratorTrainer,
    monkeypatch: MonkeyPatch,
) -> None:
    # skip validation of rag system
    monkeypatch.setenv("FEDRAG_SKIP_VALIDATION", "1")

    # no retriever trainer set
    manager = HuggingFaceRAGTrainerManager(
        mode="generator",
        generator_trainer=generator_trainer,
    )

    with pytest.raises(
        UnspecifiedRetrieverTrainer,
        match="Attempted to perform retriever trainer with an unspecified trainer.",
    ):
        manager._train_retriever()


def test_private_train_generator_raises_unspecified_generator_error(
    retriever_trainer: BaseGeneratorTrainer,
    monkeypatch: MonkeyPatch,
) -> None:
    # skip validation of rag system
    monkeypatch.setenv("FEDRAG_SKIP_VALIDATION", "1")

    # no retriever trainer set
    manager = HuggingFaceRAGTrainerManager(
        mode="retriever",
        retriever_trainer=retriever_trainer,
    )

    with pytest.raises(
        UnspecifiedGeneratorTrainer,
        match="Attempted to perform generator trainer with an unspecified trainer.",
    ):
        manager._train_generator()
